import json
import os
import re
import subprocess
import sys
import tempfile
import threading
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Optional, List, Dict

from requests import HTTPError

from . import JobData
from ..client.nexus import NexusClient
from ..client.rest_client_factory import RestClientFactory
from ..deployment.jobs.airflow import AirflowJobDeployment
from ..deployment.jobs.databricks import DatabricksJobsDeployment, get_fabric_label
from ..entities.project import Project, is_cross_project_pipeline
from ..utility import custom_print as log, Either, is_online_mode
from ..utils.constants import SCALA_LANGUAGE
from ..utils.exceptions import ProjectBuildFailedException
from ..utils.project_config import ProjectConfig, EMRInfo, DataprocInfo, DeploymentMode
from ..utils.project_models import StepMetadata, Operation, StepType, Status, LogLevel, Colors


def get_package_name(project_language: str, pipeline_name: str):
    if project_language == SCALA_LANGUAGE:
        return f'{pipeline_name}.jar'
    else:
        # todo combine in a single regex
        regex_match = r"[^\w\d.]+"
        underscore_regex = r"(_)\1+"
        result = re.sub(regex_match, "_", pipeline_name)
        result = re.sub(underscore_regex, "_", result)
        return f'{result}-1.0-py3-none-any.whl'


class PipelineDeployment:
    def __init__(self, project: Project, databricks_jobs: DatabricksJobsDeployment,
                 airflow_jobs: AirflowJobDeployment,
                 project_config: ProjectConfig, job_ids: Optional[List[str]] = None,
                 pipelines_to_build: Optional[List[str]] = None):

        self.job_ids = job_ids
        self.pipelines_to_build = pipelines_to_build

        self.databricks_jobs = databricks_jobs
        self.airflow_jobs = airflow_jobs

        self.project = project
        self.project_config = project_config
        self.deployment_mode = project_config.configs_override.mode
        self.are_tests_enabled = project_config.configs_override.tests_enabled

        self.pipeline_id_to_local_path = {}
        self.has_pipelines = False  # in case deployment doesn't have any pipelines.

    @property
    def _all_jobs(self) -> Dict[str, JobData]:
        return {
            **self.databricks_jobs.valid_databricks_jobs,
            **self.airflow_jobs.valid_airflow_jobs
        }

    def summary(self):

        summary = []
        for pipeline_id, pipeline_name in self._pipeline_components_from_jobs().items():
            summary.append(f"Pipeline {pipeline_id} will be build and uploaded.")
        return summary

    def headers(self):
        headers = []
        for pipeline_id, pipeline_name in self._pipeline_components_from_jobs().items():
            headers.append(StepMetadata(pipeline_id, f"Build {pipeline_name} pipeline",
                                        Operation.Build, StepType.Pipeline))
        return headers

    def _build_and_upload_online(self, pipeline_jobs):
        responses = []
        max_workers = os.getenv("PBT_MAX_WORKERS", 2)
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = {executor.submit(self.build_and_upload_pipeline, pipeline_id, pipeline_name): (
                pipeline_id, pipeline_name) for pipeline_id, pipeline_name in pipeline_jobs}

            for future in as_completed(futures):
                responses.append(future.result())
        return responses

    def _build_and_upload_offline(self, pipeline_jobs):
        responses = []
        for index, (pipeline_id, pipeline_name) in enumerate(pipeline_jobs, start=1):
            try:
                log(f"{Colors.OKBLUE}\nBuilding pipeline {pipeline_id} [{index}/{len(pipeline_jobs)}]{Colors.ENDC}\n\n",
                    indent=1)
                response = self.build_and_upload_pipeline(pipeline_id, pipeline_name)
                responses.append(response)
            except Exception as exc:
                log(f"{Colors.FAIL}Pipeline {pipeline_id} generated an exception: {exc}{Colors.ENDC}",
                    step_id=pipeline_id, indent=1)
        return responses

    def build_and_upload(self):
        use_threads = is_online_mode()

        if len(self._pipeline_components_from_jobs().items()) > 0:
            log(f"\n\n{Colors.OKCYAN}Building {len(self._pipeline_to_list_fabrics)} Pipelines {Colors.ENDC}\n")

        pipelines_from_job = self._pipeline_components_from_jobs().items()

        if use_threads and len(pipelines_from_job) > 1:
            return self._build_and_upload_online(pipelines_from_job)

        else:
            return self._build_and_upload_offline(pipelines_from_job)

    def build_and_upload_pipeline(self, pipeline_id, pipeline_name):
        log(step_id=pipeline_id, step_status=Status.RUNNING)

        all_available_fabrics = set(self.project_config.fabric_config.list_all_fabrics())
        relevant_fabrics_for_pipeline = [fabric for fabric in self._pipeline_to_list_fabrics.get(pipeline_id) if
                                         fabric in all_available_fabrics]

        pipeline_builder = PackageBuilderAndUploader(self.project, pipeline_id, pipeline_name,
                                                     self.project_config,
                                                     are_tests_enabled=self.are_tests_enabled,
                                                     fabrics=relevant_fabrics_for_pipeline)
        return pipeline_builder.build_and_upload_pipeline()

    def deploy(self):
        return self.build_and_upload()

    def validate(self, treat_warning_as_errors: bool = False):
        all_pipelines = self._pipeline_components_from_jobs()
        log(f"{Colors.OKBLUE} Validating pipelines {Colors.ENDC}")

        overall_validate_status = True

        for pipeline_id, pipeline_name in all_pipelines.items():
            log(f"\n\nValidating pipeline {pipeline_name} \n")
            rdc = self.project.load_pipeline_folder(pipeline_id)
            workflow = rdc.get('.prophecy/workflow.latest.json', None)

            num_errors = 0
            num_warnings = 0

            if workflow is None:
                log(f"\n{Colors.FAIL}Empty Pipeline Found: {pipeline_name}!{Colors.ENDC}")
            else:
                workflow_json = json.loads(workflow)
                if "diagnostics" in workflow_json:
                    diagnostics = workflow_json["diagnostics"]
                    for diagnostic in diagnostics:
                        if diagnostic.get("severity") == 1:
                            log(f"\n{Colors.FAIL} {pipeline_name}: {diagnostic.get('message')}{Colors.ENDC}")
                            num_errors += 1
                        elif diagnostic.get("severity") == 2:
                            log(f"\n{Colors.WARNING} {pipeline_name}: {diagnostic.get('message')}{Colors.ENDC}")
                            num_warnings += 1
                    log(f"\n{pipeline_name} has {num_errors} errors and {num_warnings} warnings.")
                    if num_errors > 0 or (treat_warning_as_errors and num_warnings > 0):
                        log(f"\n{Colors.FAIL}Pipeline is Broken: {pipeline_name}{Colors.ENDC}")
                        overall_validate_status = False
                else:
                    log(f"\n{Colors.OKBLUE} Pipeline is validated: {pipeline_name}{Colors.ENDC}")

        if not overall_validate_status:
            sys.exit(1)
        else:
            sys.exit(0)

    def test(self):
        log(f"{Colors.OKBLUE}Testing pipelines{Colors.ENDC}")

        responses = {}
        for pipeline_id, pipeline_name in self._pipeline_components_from_jobs().items():
            log(f"{Colors.OKGREEN} Testing pipeline `{pipeline_id}` {Colors.ENDC}", step_id=pipeline_id)
            log(step_id=pipeline_id, step_status=Status.RUNNING)

            pipeline_builder = PackageBuilderAndUploader(self.project, pipeline_id, pipeline_name,
                                                         self.project_config,
                                                         are_tests_enabled=True,
                                                         fabrics=self._pipeline_to_list_fabrics.get(pipeline_id))
            return_code = pipeline_builder.test()
            responses[pipeline_id] = return_code
        for pipeline_id, return_code in responses.items():
            if return_code in (0, 5):
                log(f"{Colors.OKGREEN} Pipeline test succeeded : `{pipeline_id}` {Colors.ENDC}")
            else:
                log(f"{Colors.FAIL} Pipeline test failed : `{pipeline_id}`  {Colors.ENDC}")

    def build(self, pipeline_names: str = "", ignore_build_errors: bool = False, ignore_parse_errors: bool = False):
        # these can be names and ids.
        all_pipeline_ids = self._pipeline_to_list_fabrics_full_deployment.keys()
        pipeline_ids_to_name = {pipeline_id: self.project.get_pipeline_name(pipeline_id) for pipeline_id in
                                all_pipeline_ids if self.project.get_pipeline_name(pipeline_id) is not None}

        if pipeline_names is not None and len(pipeline_names) > 0:
            pipelines_set = {pipeline for pipeline in pipeline_names.split(",")}
            pipeline_ids_to_name = {pipeline_id: pipeline_name for pipeline_id, pipeline_name in
                                    pipeline_ids_to_name.items()
                                    if pipeline_name in pipelines_set}

            pipelines_to_skip_build = {pipeline_id: pipeline_name for pipeline_id, pipeline_name in
                                       pipeline_ids_to_name.items()
                                       if pipeline_name not in pipelines_set}

            if len(pipelines_to_skip_build) > 0:
                log(f"{Colors.WARNING}Skipping build for pipelines {pipelines_to_skip_build}, Please check the ids/names of provided pipelines {Colors.ENDC}")

        log(f"\n\n{Colors.OKBLUE}Building pipelines {len(pipeline_ids_to_name)}{Colors.ENDC}\n")

        build_errors = False
        for index, (pipeline_id, pipeline_name) in enumerate(pipeline_ids_to_name.items(), start=1):
            log(f"\n\n{Colors.OKGREEN}Building pipeline `{pipeline_name}`:[{index}/{len(pipeline_ids_to_name)}] {Colors.ENDC}\n",
                indent=1)
            log(step_id=pipeline_id, step_status=Status.RUNNING)

            code = self.project.load_pipeline_folder(pipeline_id)

            if (len(code) == 0) and ignore_parse_errors:
                log(f"{Colors.WARNING}Pipeline `{pipeline_name}` code is not present {Colors.ENDC}", indent=2)
                pass

            pipeline_builder = PackageBuilderAndUploader(self.project, pipeline_id, pipeline_name,
                                                         self.project_config,
                                                         are_tests_enabled=False,
                                                         fabrics=self._pipeline_to_list_fabrics.get(pipeline_id))
            try:
                build_success = pipeline_builder.build(ignore_build_errors) == 0
            except ProjectBuildFailedException:
                build_success = False

            if build_success:
                build_errors = True

            if build_success:
                log(f"\n{Colors.OKGREEN}Build for pipeline `{pipeline_name}` succeeded {Colors.ENDC}", indent=1)
            elif ignore_build_errors:
                log(f"\n{Colors.WARNING}Build for pipeline `{pipeline_name}` failed {Colors.ENDC}", indent=1)
            else:
                log(f"\n{Colors.FAIL}Build for pipeline `{pipeline_name}` failed {Colors.ENDC}", indent=1)

        if build_errors:
            sys.exit(1)
        else:
            if ignore_build_errors:
                log(f"\n{Colors.WARNING}Ignoring builds Errors as --ignore-build-errors is passed.{Colors.ENDC}")

    @property
    def _pipeline_to_list_fabrics_selective_job(self) -> Dict[str, List[str]]:
        pipeline_id_to_fabrics_dict = {}

        for job_id, job_data in self._all_jobs.items():
            for entity in job_data.pipeline_and_fabric_ids:
                if entity.entity_id not in pipeline_id_to_fabrics_dict:
                    pipeline_id_to_fabrics_dict[entity.entity_id] = set()

                pipeline_id_to_fabrics_dict[entity.entity_id].add(entity.fabric_id)

        return {pipeline_id: list(fabric_ids) for pipeline_id, fabric_ids in pipeline_id_to_fabrics_dict.items()}

    @property
    def _pipeline_to_list_fabrics_full_deployment(self) -> Dict[str, List[str]]:
        pipelines_to_fabrics = {}
        for pipeline_id in self.project.pipelines:
            pipelines_to_fabrics[pipeline_id] = self.project_config.fabric_config.list_all_fabrics()

        return pipelines_to_fabrics

    @property
    def _pipeline_to_list_fabrics(self) -> Dict[str, List[str]]:
        if self.deployment_mode == DeploymentMode.SelectiveJob:
            return self._pipeline_to_list_fabrics_selective_job
        elif not is_online_mode():
            return {**self._pipeline_to_list_fabrics_full_deployment,
                    **self._pipeline_to_list_fabrics_selective_job}
        else:
            return self._pipeline_to_list_fabrics_full_deployment

    def _pipeline_components_from_jobs(self):
        pipeline_components = {}

        for pipeline_id in list(self._pipeline_to_list_fabrics.keys()):
            pipeline_name = self.project.get_pipeline_name(pipeline_id)
            pipeline_components[pipeline_id] = pipeline_name

        return pipeline_components


# look at the nexus client, download the jar in target folder or dist folder and return or
# if it's not present in nexus, then build the jar upload to nexus and return it back.
class PackageBuilderAndUploader:

    def __init__(self, project: Project, pipeline_id: str, pipeline_name: str,
                 project_config: ProjectConfig = None, are_tests_enabled: bool = False, fabrics: List = []):

        self._pipeline_id = pipeline_id
        self._pipeline_name = pipeline_name
        self._are_tests_enabled = are_tests_enabled
        self._project = project
        self._project_langauge = project.project_language
        self._base_path = project.load_pipeline_base_path(pipeline_id)
        self._project_config = project_config
        self.fabrics = fabrics
        self.pipeline_upload_manager = PipelineUploadManager(self._project, self._project_config, self._pipeline_id,
                                                             self._pipeline_name,
                                                             self.fabrics)

    def _initialize_temp_folder(self):
        rdc = self._project.load_pipeline_folder(self._pipeline_id)

        temp_dir = tempfile.mkdtemp()
        for file_name, file_content in rdc.items():
            file_path = os.path.join(temp_dir, file_name)
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
            with open(file_path, 'w') as f:
                f.write(file_content)
        self._base_path = temp_dir

    def test(self):

        if self._project_langauge == SCALA_LANGUAGE:
            return self.mvn_test()
        else:
            return self.wheel_test()

    def build(self, ignore_build_errors: bool = False):

        if self._project_langauge == SCALA_LANGUAGE:
            return self.mvn_build(ignore_build_errors)
        else:
            return self.wheel_build(ignore_build_errors)

    def build_and_upload_pipeline(self) -> Either:
        pipeline_from_nexus = self._download_pipeline_from_nexus()
        if pipeline_from_nexus is not None and pipeline_from_nexus.is_right:
            return Either(right=(self._pipeline_id, pipeline_from_nexus.right))

        if not is_online_mode() and self._project_config.skip_builds:
            log("Artifact File already exist, ignoring pipeline build", step_id=self._pipeline_id, indent=2)
            return Either(right=True)

        try:
            self._initialize_temp_folder()

            log("Initialized temp folder for building the pipeline package.", step_id=self._pipeline_id, indent=2)

            if self._project_langauge == SCALA_LANGUAGE:
                self.mvn_build()
            else:
                self.wheel_build()

            path = Project.get_pipeline_whl_or_jar(self._base_path)
            log(f"{Colors.OKGREEN}Pipeline package built successfully, with path {path}{Colors.ENDC}\n",
                step_id=self._pipeline_id, indent=2)

            if self._project_config.system_config.nexus is not None:
                log("Trying to upload pipeline package to nexus.", self._pipeline_id)
                self._uploading_to_nexus(path)

            log(f"Uploading pipeline {self._pipeline_id} from path {path}", indent=2)

            return self._upload_pipeline(path)

        except Exception as e:
            log(message=f"{Colors.FAIL}Failed to build the pipeline package.{Colors.ENDC}", exception=e,
                step_id=self._pipeline_id, indent=1)
            log(step_id=self._pipeline_id, step_status=Status.FAILED)
            return Either(left=e)

    def _upload_pipeline(self, path: str) -> Either:
        return self.pipeline_upload_manager.upload_pipeline(path)

    def _download_pipeline_from_nexus(self) -> Optional[Either]:
        if self._project_config.system_config.nexus is not None:
            log("Project has nexus configured, trying to download the pipeline package.", step_id=self._pipeline_id,
                indent=2)
            return self._download_from_nexus()
        else:
            # log("Project does not have nexus configured, building the pipeline package.", step_id=self._pipeline_id, indent=2)
            return None

    def _uploading_to_nexus(self, upload_path):
        try:
            client = NexusClient.initialize_nexus_client(self._project_config)
            client.upload_file(upload_path, self._project.project_id,
                               self._pipeline_id, self._project.release_version,
                               self._get_package_name())
            log("Pipeline uploaded to nexus.", step_id=self._pipeline_id)
        except Exception as e:
            log("Failed to upload pipeline to nexus", e, self._pipeline_id)

    def _download_from_nexus(self):
        try:
            client = NexusClient.initialize_nexus_client(self._project_config)
            package_name = self._get_package_name()
            response = client.download_file(package_name,
                                            self._project.project_id,
                                            self._project.release_version,
                                            self._pipeline_id)
            log("Pipeline downloaded from nexus.", step_id=self._pipeline_id)
            return Either(right=response)
        except Exception as e:
            log("Failed to download pipeline from nexus", e, self._pipeline_id)
            return Either(left=e)

    def _get_package_name(self):
        if self._project_langauge == SCALA_LANGUAGE:
            return f'{self._pipeline_name}.jar'
        else:
            # todo combine in a single regex
            regex_match = r"[^\w\d.]+"
            underscore_regex = r"(_)\1+"
            result = re.sub(regex_match, "_", self._pipeline_name)
            result = re.sub(underscore_regex, "_", result)
            return f'{result}-1.0-py3-none-any.whl'

    def mvn_build(self, ignore_build_errors: bool = False):
        mvn = "mvn"
        command = [mvn, "package", "-DskipTests"] if not self._are_tests_enabled else [mvn, "package",
                                                                                       "-Dfabric=default"]

        log(f"Running mvn command {command}", step_id=self._pipeline_id, indent=2)

        return self._build(command, ignore_build_errors)

    def mvn_test(self):
        mvn = "mvn"
        command = [mvn, "package", "-Dfabric=default"]
        log(f"Running mvn command {command}", step_id=self._pipeline_id, indent=2)

        return self._build(command)

    def wheel_test(self):
        separator = os.sep
        test_command = ["python3", "-m", "pytest", "-v", f"{self._base_path}{separator}test{separator}TestSuite.py"]
        log(f"Running python test {test_command}", step_id=self._pipeline_id, indent=2)
        response_code = self._build(test_command)

        return response_code

    def wheel_build(self, ignore_build_error: bool = False):
        if self._are_tests_enabled:
            response_code = self.wheel_test()

            if response_code not in (0, 5):
                raise Exception(f"Python test failed for pipeline {self._pipeline_id}")

        command = ["python3", "setup.py", "bdist_wheel"]

        log(f"Running python command {command}", step_id=self._pipeline_id, indent=2)

        return self._build(command, ignore_build_error)

    # maybe we can try another iteration with yield ?
    def _build(self, command: list, ignore_build_errors: bool = False):
        env = dict(os.environ)

        # Set the MAVEN_OPTS variable
        env["MAVEN_OPTS"] = "-Xmx1024m -XX:MaxPermSize=512m -Xss32m"

        if env.get('FABRIC_NAME', None) is None:
            env["FABRIC_NAME"] = "default"  # for python test runs.

        log(f"Running command {command} on path {self._base_path}",
            step_id=self._pipeline_id, indent=2)
        process = subprocess.Popen(command, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env,
                                   cwd=self._base_path)

        def log_output(pipe, log_function):
            while True:
                # Read line from stdout or stderr, break if EOF
                output = pipe.readline()
                if process.poll() is not None and not output:
                    break
                # Decode line and print it
                response = output.decode().strip()

                # stripping unnecessary logs
                if not re.search(r'Progress \(\d+\):', response) and len(response) != 0 and response != '\n':
                    log_function(response)

        # Create threads to read and log stdout and stderr simultaneously
        stdout_thread = threading.Thread(target=log_output,
                                         args=(
                                             process.stdout, lambda msg: log(msg, step_id=self._pipeline_id, indent=2)))
        stderr_thread = threading.Thread(target=log_output,
                                         args=(
                                             process.stderr, lambda msg: log(msg, step_id=self._pipeline_id, indent=2)))

        # Start threads
        stdout_thread.start()
        stderr_thread.start()

        # Wait for both threads to finish
        stdout_thread.join()
        stderr_thread.join()

        # Get the exit code
        return_code = process.wait()

        if return_code in (0, 5):
            log(f"Build was successful with exit code {return_code}", step_id=self._pipeline_id, indent=2)
        elif ignore_build_errors:
            log(f"Build failed with exit code {return_code}", step_id=self._pipeline_id, indent=2)
        else:
            log(f"Build failed with exit code {return_code}", step_id=self._pipeline_id, indent=2)
            raise ProjectBuildFailedException(f"Build failed with exit code {return_code}")

        return return_code


class PipelineUploader(ABC):

    @abstractmethod
    def upload_pipeline(self, path: str):
        pass

    def exists(self) -> bool:
        pass


class PipelineUploadManager(PipelineUploader, ABC):
    def __init__(self, project: Project, project_config: ProjectConfig, pipeline_id: str, pipeline_name: str,
                 all_fabrics: List[str]):
        self.project = project
        self.project_config = project_config
        self.pipeline_id = pipeline_id
        self.pipeline_name = pipeline_name
        self.all_fabrics = all_fabrics

    def upload_pipeline(self, from_path: str):
        try:
            if from_path is None:
                raise Exception(f"Pipeline build failed {self.pipeline_id}")

            file_name_with_extension = os.path.basename(from_path)

            if file_name_with_extension.endswith("-1.0.jar"):
                # scala based pipeline
                file_name = file_name_with_extension.replace("-1.0.jar", ".jar")
            else:
                # python based pipeline they are correctly generated.
                file_name = file_name_with_extension

            subscribed_project_id, subscribed_project_release_version, path = is_cross_project_pipeline(
                self.pipeline_id)

            if subscribed_project_id is not None:
                to_path = f"{subscribed_project_id}/{subscribed_project_release_version}"
            else:
                to_path = f"{self.project.project_id}/{self.project.release_version}"

            responses = []

            for fabric_id in self.all_fabrics:
                try:
                    fabric_info = self.project_config.fabric_config.get_fabric(fabric_id)
                    fabric_name = fabric_info.name
                    db_info = fabric_info.databricks
                    emr_info = fabric_info.emr
                    dataproc_info = fabric_info.dataproc

                    if db_info is not None:
                        pipeline_uploader = DatabricksPipelineUploader(self.project, self.project_config,
                                                                       self.pipeline_id, to_path, from_path,
                                                                       file_name,
                                                                       fabric_id, fabric_name)

                    elif emr_info is not None:
                        pipeline_uploader = EMRPipelineUploader(self.project, self.project_config,
                                                                self.pipeline_id, from_path, to_path,
                                                                file_name, fabric_id, fabric_name, emr_info)

                    elif dataproc_info is not None:
                        pipeline_uploader = DataprocPipelineUploader(self.project, self.project_config,
                                                                     self.pipeline_id, from_path, to_path,
                                                                     file_name, fabric_id, fabric_name, dataproc_info)
                    else:
                        log(f"{Colors.WARNING}Fabric id `{fabric_id}` with name `{fabric_name}` is not supported for pipeline upload{Colors.ENDC}",
                            step_id=self.pipeline_id, )
                        pipeline_uploader = DummyPipelineUploader()

                    responses.append(pipeline_uploader.upload_pipeline(""))

                except Exception as e:
                    log(f"{Colors.FAIL}Error while uploading pipeline {self.pipeline_id} for fabric {fabric_id}{Colors.ENDC}",
                        step_id=self.pipeline_id, exception=e, level=LogLevel.TRACE, indent=2)
                    log(step_status=Status.FAILED, step_id=self.pipeline_id)
                    responses.append(Either(left=e))

            if all([response.is_right for response in responses]):
                log(step_status=Status.SUCCEEDED, step_id=self.pipeline_id)
                return Either(right=True)
            else:
                log(step_status=Status.FAILED, step_id=self.pipeline_id)
                return Either(left=responses)

        except Exception as e:
            log(f"{Colors.FAIL}Error while uploading pipeline {self.pipeline_id}{Colors.ENDC}",
                step_id=self.pipeline_id,
                exception=e, indent=2)
            log(step_status=Status.FAILED, step_id=self.pipeline_id)
            return Either(left=e)

    def exists(self) -> bool:
        response = True
        file_name = get_package_name(self.project.project_language, self.pipeline_name)

        subscribed_project_id, subscribed_project_release_version, path = self.project.is_cross_project_pipeline(
            self.pipeline_id)

        if subscribed_project_id is not None:
            to_path = f"{subscribed_project_id}/{subscribed_project_release_version}"
        else:
            to_path = f"{self.project.project_id}/{self.project.release_version}"

        for fabric_id in self.all_fabrics:
            fabric_info = self.project_config.fabric_config.get_fabric(fabric_id)
            fabric_name = fabric_info.name
            db_info = fabric_info.databricks
            emr_info = fabric_info.emr
            dataproc_info = fabric_info.dataproc

            if db_info is not None:
                pipeline_uploader = DatabricksPipelineUploader(self.project, self.project_config,
                                                               self.pipeline_id, to_path, "", file_name,
                                                               fabric_id, fabric_name)

            elif emr_info is not None:
                pipeline_uploader = EMRPipelineUploader(self.project, self.project_config,
                                                        self.pipeline_id, to_path, "", file_name, fabric_id,
                                                        fabric_name, emr_info)

            elif dataproc_info is not None:
                pipeline_uploader = DataprocPipelineUploader(self.project, self.project_config,
                                                             self.pipeline_id, to_path, "", file_name,
                                                             fabric_id, fabric_name, dataproc_info)
            else:
                pipeline_uploader = DummyPipelineUploader()

            response = response and pipeline_uploader.exists()

        return response


class EMRPipelineUploader(PipelineUploader, ABC):
    def __init__(self, project: Project, project_config: ProjectConfig, pipeline_id: str,
                 from_path: str, to_path: str, file_name: str, fabric_id: str, fabric_name: str, emr_info: EMRInfo):
        self.project = project
        self.project_config = project_config
        self.pipeline_id = pipeline_id

        self.from_path = from_path
        self.to_path = to_path
        self.file_name = file_name
        self.fabric_id = fabric_id
        self.fabric_name = fabric_name

        self.emr_info = emr_info
        self.file_name = file_name

        self.rest_client_factory = RestClientFactory.get_instance(RestClientFactory, project_config.fabric_config)
        self.base_path = self.project_config.system_config.get_s3_base_path()
        self.upload_path = f"{self.emr_info.bare_path_prefix()}/{self.base_path}/{self.to_path}/pipeline/{self.file_name}".lstrip(
            "/")

    def upload_pipeline(self, path: str):

        try:
            client = self.rest_client_factory.s3_client(self.fabric_id)
            client.upload_file(self.emr_info.bare_bucket(), self.upload_path, self.from_path)

            log(f"{Colors.OKGREEN}Uploaded pipeline to s3, from path {self.from_path} to path {self.upload_path} for fabric {self.fabric_name}{Colors.ENDC}",
                step_id=self.pipeline_id, indent=2)

            if self.project.project_language == "python":
                content = self.project.get_py_pipeline_main_file(self.pipeline_id)
                pipeline_name = self.pipeline_id.split("/")[1]
                launcher_path = f"{self.emr_info.bare_path_prefix()}/{self.base_path}/{self.to_path}/pipeline/{pipeline_name}/launcher.py".lstrip(
                    '/')
                client.upload_content(self.emr_info.bare_bucket(), launcher_path, content)

                log(f"{Colors.OKGREEN}Uploading py pipeline launcher to to-path {launcher_path} for fabric {self.fabric_name}{Colors.ENDC}",
                    step_id=self.pipeline_id, indent=2)
            return Either(right=True)
        except Exception as e:
            log(f"{Colors.WARNING}Unknown Exception while uploading pipeline to emr, from-path {self.from_path} to path {self.upload_path} for fabric {self.fabric_name}, Ignoring{Colors.ENDC}",
                exception=e, step_id=self.pipeline_id, indent=2)
            return Either(right=True)

    def exists(self) -> bool:
        return False


class DatabricksPipelineUploader(PipelineUploader, ABC):
    def __init__(self, project: Project, project_config: ProjectConfig, pipeline_id: str,
                 to_path: str, file_path: str, file_name: str, fabric_id: str, fabric_name: str):
        self.project = project
        self.project_config = project_config
        self.file_name = file_name
        self.file_path = file_path
        self.pipeline_id = pipeline_id
        self.to_path = to_path
        self.fabric_id = fabric_id
        self.fabric_name = fabric_name
        self.fabric_label = get_fabric_label(fabric_name, fabric_id)

        self.rest_client_factory = RestClientFactory.get_instance(RestClientFactory, project_config.fabric_config)
        self.base_path = self.project_config.system_config.get_dbfs_base_path()
        self.upload_path = f"{self.base_path}/{self.to_path}/pipeline/{self.file_name}"

    def upload_pipeline(self, path: str) -> Either:
        try:
            client = self.rest_client_factory.databricks_client(self.fabric_id)
            client.upload_src_path(self.file_path, self.upload_path)
            log(f"Uploading pipeline to databricks from-path `{self.file_path}` to path `{self.upload_path}` for fabric `{self.fabric_label}`",
                step_id=self.pipeline_id, indent=2)

            return Either(right=True)

        except HTTPError as e:
            response = e.response.content.decode('utf-8')
            log(step_id=self.pipeline_id, message=response)

            if e.response.status_code == 401 or e.response.status_code == 403:
                log(f'{Colors.WARNING}Error on uploading pipeline to databricks from path {self.file_path} to path {self.upload_path} for fabric {self.fabric_label}, but ignoring{Colors.ENDC}',
                    exception=e, step_id=self.pipeline_id, indent=2)
                return Either(right=True)
            else:
                log(f"{Colors.FAIL}HttpError on uploading pipeline to databricks from-path {self.file_path} to path {self.upload_path} for fabric {self.fabric_label}{Colors.ENDC}",
                    exception=e, step_id=self.pipeline_id, indent=2)
                return Either(left=e)

        except Exception as e:
            log(f"{Colors.FAIL}Unknown Exception on uploading pipeline to databricks from-path {self.file_path} to path {self.upload_path} for fabric {self.fabric_label}, ignoring exception{Colors.ENDC}",
                exception=e, step_id=self.pipeline_id, indent=2)
            return Either(right=True)

    def exists(self) -> bool:
        try:
            log(f"Checking if path {self.upload_path} already exists.", self.pipeline_id, indent=2)
            client = self.rest_client_factory.databricks_client(self.fabric_id)
            return client.path_exist(self.upload_path)
        except Exception as e:
            log(f"{Colors.WARNING} Failed checking path {self.upload_path}{Colors.ENDC}", step_id=self.pipeline_id,
                exception=e,
                indent=2)
            return False


class DummyPipelineUploader(PipelineUploader, ABC):

    def upload_pipeline(self, path: str):
        return Either(right=True)

    def exists(self) -> bool:
        return True


class DataprocPipelineUploader(PipelineUploader, ABC):
    def __init__(self, project: Project, project_config: ProjectConfig, pipeline_id: str,
                 from_path: str, to_path: str, file_name: str, fabric_id: str, fabric_name: str,
                 dataproc_info: DataprocInfo):
        self.project = project
        self.project_config = project_config
        self.pipeline_id = pipeline_id

        self.from_path = from_path
        self.to_path = to_path
        self.fabric_id = fabric_id
        self.fabric_name = fabric_name
        self.fabric_label = get_fabric_label(fabric_name, fabric_id)

        self.dataproc_info = dataproc_info
        self.file_name = file_name
        self.rest_client_factory = RestClientFactory.get_instance(RestClientFactory, project_config.fabric_config)
        self.base_path = self.project_config.system_config.get_s3_base_path()
        self.upload_path = f"{self.dataproc_info.bare_path_prefix()}/{self.base_path}/{self.to_path}/pipeline/{self.file_name}".lstrip(
            "/")

    def upload_pipeline(self, path: str):
        try:
            client = self.rest_client_factory.dataproc_client(self.fabric_name)
            client.put_object_from_file(self.dataproc_info.bare_bucket(), self.upload_path, self.from_path)

            log(f"{Colors.OKGREEN}Uploaded pipeline to data-proc, from-path {self.from_path} to to-path {self.upload_path} for fabric {self.fabric_label}{Colors.ENDC}",
                step_id=self.pipeline_id, indent=2)

            if self.project.project_language == "python":
                content = self.project.get_py_pipeline_main_file(self.pipeline_id)
                pipeline_name = self.pipeline_id.split("/")[1]
                launcher_path = f"{self.dataproc_info.bare_path_prefix()}/{self.base_path}/{self.to_path}/pipeline/{pipeline_name}/launcher.py".lstrip(
                    '/')
                client.put_object(self.dataproc_info.bare_bucket(), launcher_path, content)

                log(f"{Colors.OKGREEN}Uploading py pipeline launcher to to-path {launcher_path} and bucket {self.dataproc_info.bare_bucket()} for fabric {self.fabric_label}{Colors.ENDC}",
                    step_id=self.pipeline_id, indent=2)
            return Either(right=True)

        except Exception as e:
            log(f"{Colors.WARNING}Unknown Exception while uploading pipeline to data-proc, from-path {self.from_path} to to-path {self.upload_path} for fabric {self.fabric_label}, ignoring exception{Colors.ENDC}",
                exception=e, step_id=self.pipeline_id, indent=2)
            return Either(right=True)

    def exists(self) -> bool:
        return True



import json
import os
import re
from abc import ABC
from typing import Optional, List

import yaml

from ..utility import custom_print as log
from ..utils.constants import PBT_FILE_NAME, LANGUAGE, JOBS, PIPELINES, \
    PIPELINE_CONFIGURATIONS, CONFIGURATIONS, JSON_EXTENSION, BASE_PIPELINE, PROJECT_ID_PLACEHOLDER_REGEX, \
    PROJECT_RELEASE_VERSION_PLACEHOLDER_REGEX, PROJECT_RELEASE_TAG_PLACEHOLDER_REGEX, GEMS, \
    PROJECT_URL_PLACEHOLDER_REGEX
from ..utils.exceptions import ProjectPathNotFoundException, ProjectFileNotFoundException

SUBSCRIBED_ENTITY_URI_REGEX = re.compile(
    r"gitUri=(.*)&subPath=(.*)&tag=(.*)&projectSubscriptionProjectId=(.*)&path=(.*)")


def is_cross_project_pipeline(pipeline):
    match = SUBSCRIBED_ENTITY_URI_REGEX.search(pipeline)

    if match:
        git_uri, sub_path, tag, project_subscription_project_id, path = match.groups()
        return project_subscription_project_id, tag.split('/')[1], path
    else:
        return None, None, None


def _read_file_content(file_path: str) -> Optional[str]:
    if not os.path.exists(file_path):
        return None
        # If the file exists, read its content
    with open(file_path, 'r') as file:
        content = file.read()
        return content


class Project:
    _DATABRICKS_JOB_JSON = "databricks-job.json"
    _CODE_FOLDER = "code"

    def __init__(self, project_path: str,
                 project_id: Optional[str] = None,
                 release_tag: Optional[str] = None,
                 release_version: Optional[str] = None,
                 dependant_project_list: Optional[str] = None):

        self.project_id = project_id
        self.project_path = project_path

        self.release_tag = release_tag
        self.release_version = release_version
        self.md_url = os.environ.get("MD_PATH", "")

        self.pbt_project_dict = {}
        self.project_language = None
        self.jobs = {}
        self.pipelines = {}
        self.pipeline_id_to_name = {}
        self.gems = {}

        self._verify_project()
        self._load_project_config()

        self._extract_project_info()
        self.pipeline_configurations = self._load_pipeline_configurations()

        log(f"Project name {self.pbt_project_dict.get('name', '')}")
        pipelines_str = ", ".join(
            map(
                lambda pipeline: "%s (%s)" % (pipeline["name"], pipeline["language"]),
                self.pipelines.values(),
            )
        )

        jobs_str = ", ".join(
            map(
                lambda job: "%s" % (job["name"]),
                self.jobs.values(),
            )
        )
        log(f"Found {len(self.jobs)} jobs: {jobs_str}")
        log(f"Found {len(self.pipelines)} pipelines: {pipelines_str}\n  ")

        self.dependant_project = None

        if dependant_project_list is not None and len(dependant_project_list) > 0:
            project_paths = dependant_project_list.split(",")
            dependant_projects = []
            log("Parsing dependent projects")
            for path in project_paths:
                dependant_projects.append(Project(path, ""))

            self.dependant_project = DependentProjectCli(dependant_projects)
        else:
            self.dependant_project = DependentProjectAPP(project_path, self)

    @property
    def name(self) -> str:
        return self.pbt_project_dict.get('name')

    def load_databricks_job(self, job_id: str) -> Optional[str]:
        try:
            path = os.path.join(self.project_path, job_id, self._CODE_FOLDER, self._DATABRICKS_JOB_JSON)
            content = _read_file_content(path)

            if content is not None:
                return self._replace_placeholders(self._DATABRICKS_JOB_JSON, content)

            return content
        except Exception:
            return None

    def load_airflow_base_folder_path(self, job_id):
        return os.path.join(self.project_path, job_id, self._CODE_FOLDER)

    def load_airflow_folder(self, job_id):
        return {path: self._replace_placeholders(path, content)
                for path, content in
                self._read_directory(os.path.join(self.project_path, job_id, self._CODE_FOLDER)).items()
                }

    def load_airflow_folder_with_placeholder(self, job_id):
        return self._read_directory(os.path.join(self.project_path, job_id, self._CODE_FOLDER))

    def load_airflow_aspect(self, job_id: str) -> Optional[str]:
        path = os.path.join(self.project_path, job_id, "pbt_aspects.yml")
        return _read_file_content(path)

    def load_pipeline_base_path(self, pipeline):
        return os.path.join(self.project_path, pipeline, self._CODE_FOLDER)

    def load_pipeline_folder(self, pipeline):
        base_path = os.path.join(self.project_path, pipeline, self._CODE_FOLDER)

        content = self._read_directory(base_path)
        if len(content) > 0:
            return content
        else:
            return self.dependant_project.load_pipeline_folder(pipeline)

    @staticmethod
    def get_pipeline_whl_or_jar(base_path: str):
        for dir_path, dir_names, filenames in os.walk(base_path):
            for filename in filenames:
                full_path = os.path.join(dir_path, filename)
                if full_path.endswith(".whl") or full_path.endswith(".jar"):
                    return full_path

    def get_py_pipeline_main_file(self, pipeline_id):
        try:
            main_file = os.path.join(self.project_path, pipeline_id, self._CODE_FOLDER, "main.py")
            data = _read_file_content(main_file)
            return data
        except Exception:
            return self.dependant_project.get_py_pipeline_main_file(pipeline_id)

    def get_pipeline_absolute_path(self, pipeline_id):
        return os.path.join(os.path.join(self.project_path, pipeline_id), "code")

    def get_pipeline_id(self, pipeline_name):
        return next((k for k, v in self.pipelines.items() if v['name'] == pipeline_name), None)

    def get_pipeline_name(self, pipeline_id):
        try:
            pipeline_path = pipeline_id
            pipelines = self.pipelines
            return pipelines.get(pipeline_path, {}).get('name', pipeline_path.split('/')[-1])
        except Exception:
            return self.dependant_project.get_pipeline_name(pipeline_id)

    def _read_directory(self, base_path: str):
        rdc = {}

        for dir_path, dir_names, filenames in os.walk(base_path):

            # Add resources to RDC
            if 'resources' in dir_names:
                resource_dir_path = os.path.join(dir_path, 'resources')
                for resource_subdir_path, _, resource_filenames in os.walk(resource_dir_path):
                    for resource_filename in resource_filenames:
                        resource_full_path = os.path.join(resource_subdir_path, resource_filename)
                        resource_content = _read_file_content(resource_full_path)
                        if resource_content is not None:
                            relative_path = os.path.relpath(resource_full_path, base_path)
                            rdc[relative_path] = resource_content

            # Add generated code to RDC
            for filename in filenames:
                if filename.endswith(('.py', '.json', '.conf', '.scala', 'pom.xml')):
                    full_path = os.path.join(dir_path, filename)
                    content = _read_file_content(full_path)
                    if content is not None:
                        relative_path = os.path.relpath(full_path, base_path)
                        rdc[relative_path] = content
        return rdc

    def _load_subscribed_project_yml(self, subscribed_project_id: str):
        path = os.path.join(self.project_path, ".prophecy", subscribed_project_id, "pbt_project.yml")
        content = _read_file_content(path)
        if content is not None:
            return yaml.safe_load(content)

    def _verify_project(self):
        if not os.path.exists(self.project_path):
            raise ProjectPathNotFoundException(f"Project path does not exist {self.project_path}")

        if not os.path.exists(os.path.join(self.project_path, PBT_FILE_NAME)):
            raise ProjectFileNotFoundException(f"Project file does not exist at path {self.project_path}")

    def _load_project_config(self):
        pbt_project_path = os.path.join(self.project_path, PBT_FILE_NAME)
        content = _read_file_content(pbt_project_path)
        if content is not None:
            self.pbt_project_dict = yaml.safe_load(content)

    def _extract_project_info(self):
        self.project_language = self.pbt_project_dict.get(LANGUAGE, None)
        self.jobs = self.pbt_project_dict.get(JOBS, {})
        self.pipelines = self.pbt_project_dict.get(PIPELINES, {})
        self.gems = self.pbt_project_dict.get(GEMS, {})

    def _load_pipeline_configurations(self):
        pipeline_configurations = {}

        pipeline_conf = dict(self.pbt_project_dict.get(PIPELINE_CONFIGURATIONS, []))

        for pipeline_config_path, pipeline_config_object in pipeline_conf.items():
            configurations = {}

            for configurations_key in pipeline_config_object[CONFIGURATIONS]:
                config_name = pipeline_config_object[CONFIGURATIONS][configurations_key]["name"]
                file_path = os.path.join(self.project_path, configurations_key + JSON_EXTENSION)

                config = _read_file_content(file_path)

                # in case of empty config file, we will skip it
                if config is not None:
                    configurations[config_name] = config

            pipeline_configurations[pipeline_config_object[BASE_PIPELINE]] = configurations

        return pipeline_configurations

    def _replace_placeholders(self, path: str, content: str) -> str:
        if path.endswith('.json') or path.endswith('.scala') or path.endswith('.py'):
            content = content.replace(PROJECT_ID_PLACEHOLDER_REGEX, self.project_id) \
                .replace(PROJECT_RELEASE_VERSION_PLACEHOLDER_REGEX, self.release_version) \
                .replace(PROJECT_RELEASE_TAG_PLACEHOLDER_REGEX, self.release_tag) \
                .replace(PROJECT_URL_PLACEHOLDER_REGEX, self.md_url)

        return content

    # only check for non-empty gems directory
    def non_empty_gems_directory(self):
        return os.path.exists(os.path.join(self.project_path, "gems"))

    def get_package_path(self, _pipeline_id) -> (str, str):
        pipeline_path = os.path.join(self.project_path, _pipeline_id, "code")
        for root, dirs, files in os.walk(pipeline_path):
            for file in files:
                if file.endswith(".jar") or file.endswith(".whl"):
                    return root, file

    def fabrics(self) -> List[str]:
        return list(set([content.get('fabricUID', None) for content in self.jobs.values() if
                         content.get('fabricUID', None) is not None]))

    def _stripPrefix(self, value, prefix):
        if value.startswith(prefix):
            return value[len(prefix):]
        return value

    def _find_path(self):
        for job_id, content in self.jobs.items():
            db_jobs = self.load_databricks_job(job_id)
            if db_jobs is not None:
                json_content = json.loads(db_jobs)
                for component in json_content['components']:
                    for c_name, c_content in component.items():
                        path = c_content['path']

                        if path.startswith("dbfs:/FileStore/prophecy/artifacts/"):
                            return path

        return None

    def find_customer_name(self) -> str:
        path = self._find_path()
        if path is not None:
            return path.split("/")[4]
        else:
            return "dev"

    def find_control_plane_name(self):
        path = self._find_path()
        if path is not None:
            return path.split("/")[5]
        else:
            return "execution"

    def load_main_file(self, path):
        main_file = os.path.join(self.project_path, path, self._CODE_FOLDER, "main.py")
        return _read_file_content(main_file)

    def is_cross_project_pipeline(self, from_path):
        if from_path in self.pipelines:
            return None, None, None
        return self.dependant_project.is_cross_project_pipeline(from_path)


class DependentProject(ABC):

    def is_cross_project_pipeline(self, pipeline_id):
        (pid, tag, path) = is_cross_project_pipeline(pipeline_id)
        if pid is not None and tag is not None and path is not None:
            return True
        else:
            return False

    def get_py_pipeline_main_file(self, pipeline_id) -> Optional[str]:
        pass

    def get_pipeline_name(self, pipeline_id: str) -> Optional[str]:
        pass

    def load_pipeline_folder(self, pipeline_id: str):
        pass


class DependentProjectCli(DependentProject, ABC):
    def __init__(self, dependant_projects: List[Project]):
        self.dependant_projects = dependant_projects

    def get_py_pipeline_main_file(self, pipeline_id) -> Optional[str]:
        (pid, tag, path) = is_cross_project_pipeline(pipeline_id)
        for project in self.dependant_projects:
            if path in project.pipelines:
                return project.get_py_pipeline_main_file(path)

        return None

    def get_pipeline_name(self, pipeline_id: str) -> Optional[str]:
        (pid, tag, path) = is_cross_project_pipeline(pipeline_id)
        for project in self.dependant_projects:
            if path in project.pipelines:
                return project.pipelines[path]['name']

        return None

    def load_pipeline_folder(self, pipeline_id: str):
        (pid, tag, path) = is_cross_project_pipeline(pipeline_id)
        for project in self.dependant_projects:
            if path in project.pipelines:
                return project.load_pipeline_folder(path)

        return {}


class DependentProjectAPP(DependentProject, ABC):
    def __init__(self, path: str, main_project: Project):
        self.path = path
        self.project = main_project
        self.base_path = self.project.project_path

    def get_py_pipeline_main_file(self, pipeline_id) -> Optional[str]:
        (pid, tag, path) = is_cross_project_pipeline(pipeline_id)
        if pid is not None:
            base_path = "{}/.prophecy/{}/{}".format(self.base_path, pid, path)
            return self.project.load_main_file(base_path)
        else:
            return None

    def get_pipeline_name(self, pipeline_id: str) -> Optional[str]:
        (pid, tag, path) = is_cross_project_pipeline(pipeline_id)
        if pid is not None:
            base_path = "{}/.prophecy/{}".format(self.base_path, pid)
            return Project(base_path, pid).get_pipeline_name(path)
        else:
            return None

    def load_pipeline_folder(self, pipeline_id: str):
        (pid, tag, path) = is_cross_project_pipeline(pipeline_id)
        if pid is not None:
            base_path = "{}/.prophecy/{}".format(self.base_path, pid)
            return Project(base_path, pid).load_pipeline_folder(path)
        else:
            return {}



from abc import ABC, abstractmethod
from enum import Enum
from typing import Optional, List, Dict


class OperationType(Enum):
    CREATED = 1
    UPDATED = 2
    DELETED = 3
    REFRESH = 4


class JobInfoAndOperation:
    def __init__(self, job_info, operation: OperationType):
        self.job_info = job_info
        self.operation = operation


class EntityIdToFabricId:
    def __init__(self, entity_id: str, fabric_id: str):
        self.entity_id = entity_id
        self.fabric_id = fabric_id


def invert_entity_to_fabric_mapping(entity_id_dict: Dict[str, List[EntityIdToFabricId]]) -> \
        Dict[str, List[EntityIdToFabricId]]:
    result = {}

    for entity_uri, list_of_entities in entity_id_dict.items():
        for inner in list_of_entities:
            if inner.entity_id not in result:
                result[inner.entity_id] = []
            result[inner.entity_id].append(EntityIdToFabricId(entity_uri, inner.fabric_id))

    return result


# creating an abstract class helps to morge both airflow and databricks jobs and have common behavior for both of them.
class JobData(ABC):

    @property
    @abstractmethod
    def name(self) -> Optional[str]:
        pass

    @property
    @abstractmethod
    def is_valid_job(self) -> bool:
        pass

    # we can't use pbt file because it doesn't have fabric per pipeline which airflow jobs supports
    @property
    @abstractmethod
    def pipeline_and_fabric_ids(self) -> List[EntityIdToFabricId]:
        pass

    @abstractmethod
    def validate_prophecy_managed_checksum(self, salt: str) -> bool:
        pass

    @property
    @abstractmethod
    def job_files(self):
        pass

    @property
    @abstractmethod
    def dag_name(self) -> Optional[str]:
        pass

    @property
    @abstractmethod
    def is_disabled(self) -> bool:
        pass

    @property
    @abstractmethod
    def is_enabled(self) -> bool:
        pass

    @property
    @abstractmethod
    def fabric_id(self) -> Optional[str]:
        pass

    @property
    @abstractmethod
    def pipelines(self) -> List[str]:
        pass

    @property
    @abstractmethod
    def has_dbt_component(self) -> bool:
        pass



import boto3
from botocore.client import BaseClient


class S3Client:

    def __init__(self, region, access_key, secret_key, aws_session_token=None):
        self.s3: BaseClient = boto3.client(
            "s3",
            region_name=region,
            aws_access_key_id=access_key,
            aws_secret_access_key=secret_key,
            aws_session_token=aws_session_token
        )

    def upload_file(self, bucket, key, file_path):
        self.s3.upload_file(file_path, bucket, key)

    def upload_content(self, bucket, path, content):

        self.s3.put_object(Bucket=bucket, Key=path, Body=content)



from click.testing import CliRunner
from src.pbt import test
import os

PROJECT_PATH = str(os.getcwd()) + "/test/resources/HelloWorld"


def test_test_path_default():
    runner = CliRunner()
    result = runner.invoke(test, ["--path", PROJECT_PATH])
    print(result.output)
    assert "Found 2 jobs: test-job1234, job-another" in result.output
    assert (
            "Found 4 pipelines: customers_orders1243 (python), report_top_customers (python),\njoin_agg_sort (python), "
            "farmers-markets-irs (python)" in result.output
    )
    assert "Testing All pipelines" in result.output
    assert "Unit Testing pipeline pipelines/customers_orders" in result.output
    assert "Unit Testing pipeline pipelines/report_top_customers" in result.output
    assert "Unit Testing pipeline pipelines/join_agg_sort" in result.output
    assert "Unit Testing pipeline pipelines/farmers-markets-irs" in result.output


def test_test_with_pipeline_filter():
    runner = CliRunner()
    result = runner.invoke(test, ["--path", PROJECT_PATH, '--pipelines', 'report_top_customers,join_agg_sort'])
    print(result.output)
    assert "Found 2 jobs: test-job1234, job-another" in result.output
    assert (
            "Found 4 pipelines: customers_orders1243 (python), report_top_customers (python),\njoin_agg_sort (python), "
            "farmers-markets-irs (python)" in result.output
    )
    assert "Pipeline Filters passed [2]: ['report_top_customers', 'join_agg_sort']" in result.output
    assert "Unit Testing pipeline pipelines/customers_orders" not in result.output
    assert "Unit Testing pipeline pipelines/report_top_customers" in result.output
    assert "Unit Testing pipeline pipelines/join_agg_sort" in result.output
    assert "Unit Testing pipeline pipelines/farmers-markets-irs" not in result.output


def test_test_with_pipeline_filter_one_notfound_pipeline():
    runner = CliRunner()
    result = runner.invoke(test, ["--path", PROJECT_PATH, '--pipelines', 'report_top_customers,notfound'])
    print(result.output)
    assert "Pipeline Filters passed [2]: ['report_top_customers', 'notfound']" in result.output
    assert "Pipelines found [1]" in result.output
    assert "Filtered pipelines doesn't match with passed filter" in result.output


def test_test_with_pipeline_filter_all_notfound_pipelines():
    runner = CliRunner()
    result = runner.invoke(test, ["--path", PROJECT_PATH, '--pipelines', 'notfound1,notfound2,notfound3'])
    print(result.output)
    assert "Pipeline Filters passed [3]: ['notfound1', 'notfound2', 'notfound3']" in result.output
    assert "Pipelines found [0]" in result.output
    assert "Filtered pipelines doesn't match with passed filter" in result.output



import collections
import json
import os
import re
import subprocess
import sys
from glob import glob
from os import listdir
from os.path import basename, isfile, join, dirname
from typing import Dict

import yaml
from databricks_cli.configure.config import _get_api_client
from databricks_cli.configure.provider import EnvironmentVariableConfigProvider
from databricks_cli.sdk import DbfsService, JobsService
from requests import HTTPError
from rich import print

from .process import Process
import tempfile


class ProphecyBuildTool:
    def __init__(
        self,
        path_root: str,
        dependent_projects_path: str = "",
        release_version: str = "",
        project_id: str = "",
        prophecy_url: str = "",
        ignore_parse_errors: bool = False,
    ):
        if not path_root:
            self._error("Path of project not passed as argument using --path.")
        self.operating_system = sys.platform
        self.path_root = path_root
        self.path_project = os.path.join(self.path_root, "pbt_project.yml")

        self._verify_project()
        self._parse_project(ignore_parse_errors)
        self.dependent_projects = {}
        if dependent_projects_path:
            print("\nParsing dependent projects")
            other_pbt_projects = glob(
                f"{dependent_projects_path}{os.sep}**{os.sep}pbt_project.yml",
                recursive=True,
            )
            for dependent_project in other_pbt_projects:
                self.dependent_projects[dirname(dependent_project)] = ProphecyBuildTool(dirname(dependent_project))
        self.dbfs_service = None
        self.jobs_service = None
        self.api_client = None
        self.python_cmd, self.pip_cmd = self.get_python_commands(path_root)
        self.pipelines_build_path = {}
        self.dependent_pipelines_build_path = {}
        self.uploaded_target_paths = set()
        self.project_release = (
            release_version
            if release_version
            else os.getenv(
                "PROJECT_RELEASE_VERSION_PLACEHOLDER",
                "__PROJECT_RELEASE_VERSION_PLACEHOLDER__",
            )
        )
        self.project_id = (
            project_id if project_id else os.getenv("PROJECT_ID_PLACEHOLDER", "__PROJECT_ID_PLACEHOLDER__")
        )
        self.prophecy_url = (
            prophecy_url if prophecy_url else os.getenv("PROPHECY_URL_PLACEHOLDER", "__PROPHECY_URL_PLACEHOLDER__")
        )

    def get_databricks_job_json_path(self, path_job):
        path_job_absolute = os.path.join(os.path.join(self.path_root, path_job), "code")
        return os.path.join(path_job_absolute, "databricks-job.json")

    def get_prophecy_job_json_path(self, path_job):
        path_job_absolute = os.path.join(os.path.join(self.path_root, path_job), "code")
        return os.path.join(path_job_absolute, "prophecy-job.json")

    # find all the pipelines which are needed to be build for the jobs
    def generate_pipeline_deps(self):
        all_jobs_deps = []
        for job_idx, (path_job, job) in enumerate(self.jobs.items()):
            prophecy_job_definition = self.get_prophecy_job_json_path(path_job)
            with open(prophecy_job_definition, "r") as _in:
                job_definition = json.load(_in)

            pipeline_deps = [
                x["properties"]["pipelineId"]
                for x in job_definition["processes"].values()
                if x["component"] == "Pipeline"
            ]
            all_jobs_deps.extend(pipeline_deps)

        unique_deps = set(all_jobs_deps)
        print(
            f"\n[bold blue][INFO]: Total Unique pipelines dependencies found: {len(unique_deps)}\n {unique_deps}[/bold blue]"
        )
        pipelines_to_build = {k: v for k, v in self.pipelines.items() if k in unique_deps}
        print(f"\n[INFO]: Total Filtered pipelines to build: {len(pipelines_to_build)}")
        return pipelines_to_build

    def get_python_commands(self, cwd):
        if (
            Process.process_sequential(
                [
                    Process(
                        ["python3", "--version"],
                        cwd,
                        subprocess.DEVNULL,
                        subprocess.DEVNULL,
                        (self.operating_system == "win32"),
                    )
                ]
            )
            == 0
        ):
            return "python3", "pip3"
        elif (
            Process.process_sequential(
                [
                    Process(
                        ["python", "--version"],
                        cwd,
                        subprocess.DEVNULL,
                        subprocess.DEVNULL,
                        (self.operating_system == "win32"),
                    )
                ]
            )
            == 0
        ):
            return "python", "pip"
        else:
            print("ERROR: python not found")
            sys.exit(1)

    def validate(self, treat_warnings_as_errors: bool):
        pipelines = self.pipelines
        print("\n[bold blue]Validating %s pipelines [/bold blue]" % len(pipelines))
        overall_validate_status = True
        for pipeline_i, (path_pipeline, pipeline) in enumerate(pipelines.items()):
            print("\n  Validating pipeline %s [%s/%s]" % (path_pipeline, pipeline_i + 1, len(pipelines)))

            workflow_json_path_pipeline_absolute = os.path.join(
                os.path.join(self.path_root, path_pipeline),
                "code",
                ".prophecy",
                "workflow.latest.json",
            )
            if os.path.exists(workflow_json_path_pipeline_absolute):
                workflow = json.load(open(workflow_json_path_pipeline_absolute, "r"))
                num_errors = 0
                num_warnings = 0
                if "diagnostics" in workflow:
                    diagnostics = workflow["diagnostics"]
                    for diagnostic in diagnostics:
                        if diagnostic.get("severity") == 1:
                            print(f"\n[red]\[error] {pipeline['name']}: {diagnostic.get('message')}[/red]")
                            num_errors += 1
                        elif diagnostic.get("severity") == 2:
                            print(f"\n[yellow]\[warn] {pipeline['name']}: {diagnostic.get('message')}[/yellow]")
                            num_warnings += 1
                    print(f"\n{pipeline['name']} has {num_errors} errors and {num_warnings} warnings.")
                    if num_errors > 0 or (treat_warnings_as_errors and num_warnings > 0):
                        print(f"\n[bold red]Pipeline is Broken: {pipeline['name']}[/bold red]")
                        overall_validate_status = False
                else:
                    print(f"\n[bold blue] Pipeline is validated: {pipeline['name']}[/bold blue]")
            else:
                print(f"\n[bold red] Empty Pipeline Found: {pipeline['name']}![/bold red]")
                overall_validate_status = False

        if not overall_validate_status:
            sys.exit(1)
        else:
            sys.exit(0)

    def build(self, pipelines, exit_on_build_failure=True):
        if not pipelines:  # if pipelines not provided run for all pipelines
            pipelines = self.pipelines
        elif isinstance(pipelines, str):  # elif filter pipelines provided as string
            pipeline_filter = [x.strip() for x in pipelines.split(",")]
            print("\n[bold blue]Filtering pipelines: %s [/bold blue]" % str(pipeline_filter))
            pipelines = {k: v for k, v in self.pipelines.items() if k.split("/")[1] in pipeline_filter}
            if not pipelines:  # empty no matching pipeline found
                print("\n[bold yellow]No matching pipelines found for given pipelines names: %s" % (pipeline_filter))
                raise Exception()
        else:  # pipelines are provided
            print(f"\n[INFO]: Building given custom pipelines: {pipelines}")

        print("\n[bold blue]Building %s pipelines [/bold blue]" % len(pipelines))
        overall_build_status = True
        for pipeline_i, (path_pipeline, pipeline) in enumerate(pipelines.items()):
            print("\n  Building pipeline %s [%s/%s]" % (path_pipeline, pipeline_i + 1, len(pipelines)))

            path_pipeline_absolute = os.path.join(os.path.join(self.path_root, path_pipeline), "code")
            return_code = (
                self.build_python(path_pipeline_absolute, path_pipeline)
                if self.project_language == "python"
                else self.build_scala(path_pipeline_absolute)
            )

            self.pipelines_build_path[path_pipeline] = None
            build_file_found = False
            build_file_paths = []

            if self.project_language == "python":
                path_pipeline_dist = join(path_pipeline_absolute, "dist")
                build_file_paths = list(
                    filter(
                        lambda x: (x.endswith("py3-none-any.whl")),
                        glob(f"{path_pipeline_dist}{os.sep}*.whl"),
                    )
                )
            elif self.project_language == "scala":
                path_pipeline_dist = join(path_pipeline_absolute, "target")
                build_file_paths = list(
                    filter(
                        lambda x: ("jar-with-dependencies" not in x),
                        glob(f"{path_pipeline_dist}{os.sep}*.jar"),
                    )
                )

            if len(build_file_paths) > 0:
                build_file_found = True
                self.pipelines_build_path[path_pipeline] = {
                    "source_absolute": build_file_paths[0],
                    "source": os.path.basename(build_file_paths[0]),
                    "uploaded": False,
                }

            if return_code == 0:
                if build_file_found:
                    print("\n[bold blue] Build complete![/bold blue]")
                else:
                    print(
                        f"\n[bold red] Build completed but built target "
                        f"not found for pipeline: {pipeline['name']}![/bold red]"
                    )
                    overall_build_status = False
            else:
                print(f"\n[bold red] Build failed for pipeline: {pipeline['name']}![/bold red]")
                overall_build_status = False

        if not overall_build_status and exit_on_build_failure:
            sys.exit(1)
        else:
            if not exit_on_build_failure:
                print("\n[bold yellow] Ignoring builds Errors as --ignore-build-errors is passed [/bold yellow]")
            return overall_build_status, self.pipelines_build_path

    def deploy(self, fabric_ids: str = "", skip_builds: bool = False, job_ids=None):
        # not allowed to pass job_id and fabric_ids filter together ( as only job_id support incremental build and
        # deploy), fabric_ids filter builds all pipelines by default and then deploy after filtering
        if job_ids and fabric_ids:
            print("[ERROR]: Can't combine filters, Please pass either --fabric_ids or --job_ids")
            raise Exception()

        if job_ids and skip_builds:
            print(
                "[ERROR]: Can't skip builds for job_id filter,\nas it only builds depending pipelines ,\nPlease "
                "pass either --skip-builds or --job_id filter"
            )
            raise Exception()

        fabric_ids = list(i.strip() for i in fabric_ids.split(r",")) if fabric_ids else list()
        job_ids = list(i.strip() for i in job_ids.split(r",")) if job_ids else list()

        if not fabric_ids and not job_ids:
            print("Deploying jobs for all Fabrics")
        elif job_ids:  # job_id filter is provided
            print(f"Deploying jobs only for given Job IDs: {str(job_ids)}")
        else:
            print("Deploying jobs only for given Fabric IDs: %s" % (str(fabric_ids)))

        self._verify_databricks_configs()
        config = EnvironmentVariableConfigProvider().get_config()

        self.api_client = _get_api_client(config)

        self.dbfs_service = DbfsService(self.api_client)
        self.jobs_service = JobsService(self.api_client)

        if not skip_builds and not job_ids:  # build all pipelines
            self.build(dict())
        elif job_ids:
            filtered_jobs = {k: v for k, v in self.jobs.items() if k.split("/")[1] in job_ids}
            if len(filtered_jobs) == 0:
                print(
                    f"[ERROR]: No Job IDs matches with passed --job_id filter {str(job_ids)}\nAvailable Job IDs are: {self.jobs.keys()}"
                )
                raise Exception()
            else:
                self.jobs = filtered_jobs
                self.jobs_count = len(self.jobs)
            print("[INFO]: Generating depending pipelines for all jobs as '--job-ids' flag is passed.")
            pipelines_to_build = self.generate_pipeline_deps()
            self.build(pipelines_to_build)
        else:
            print("[SKIP]: Skipping builds for all pipelines as '--skip-builds' flag is passed.")

        print("\n[bold blue] Deploying %s jobs [/bold blue]" % self.jobs_count)

        pipelines_upload_failures = collections.defaultdict(list)
        job_update_failures = dict()

        for job_idx, (path_job, job) in enumerate(self.jobs.items()):
            pipelines_upload_failures_job = collections.defaultdict(list)
            print("\n[START]:  Deploying job %s [%s/%s]" % (path_job, job_idx + 1, self.jobs_count))

            path_job_definition = self.get_databricks_job_json_path(path_job)

            with open(path_job_definition, "r") as _in:
                data = _in.read()
                data = (
                    data.replace("__PROJECT_RELEASE_VERSION_PLACEHOLDER__", self.project_release)
                    .replace("__PROJECT_ID_PLACEHOLDER__", self.project_id)
                    .replace("__PROPHECY_URL_PLACEHOLDER__", self.prophecy_url)
                )
            with open(path_job_definition, "w") as _out:
                _out.write(data)

            with open(path_job_definition, "r") as _in:
                job_definition = json.load(_in)

            components = job_definition["components"]
            fabric_id = job_definition["fabric_id"]

            # if --fabric is passed, then only deploy jobs
            if fabric_ids:
                if fabric_id not in fabric_ids:  # Jobs for this fabric id should be skipped
                    print("[SKIP]: Job skipped as it belongs to fabric id (not passed): %s" % fabric_id)
                    continue
                else:
                    print("[DEPLOY]: Job being deployed for fabric id: %s" % fabric_id)

            generate_pipeline_config_from_pipeline_component = False

            for component in components:
                if "ScriptComponent" in component:
                    script_component = component["ScriptComponent"]
                    content = script_component["content"]
                    path = script_component["path"]
                    temp_file = tempfile.NamedTemporaryFile(delete=False)
                    temp_file.write(content.encode("ascii"))
                    temp_file.close()
                    print(f"Uploading script to path: {path}")
                    self.dbfs_service.put(path, overwrite=True, src_path=temp_file.name)
                    os.unlink(temp_file.name)
                if "PipelineComponent" in component:
                    pipeline_component = component["PipelineComponent"]
                    if "pipelineId" in pipeline_component:
                        pipeline_uri = pipeline_component["pipelineId"]
                    else:
                        pipeline_uri = pipeline_component["id"]

                    # Matches project_id/pipelines/pipeline_name or pipelines/pipeline_name
                    # group(2) should return pipelines/pipeline_name
                    uri_pattern = "([0-9]*)(pipelines/[-_.A-Za-z0-9 /]+)"
                    pipeline_id = re.search(uri_pattern, pipeline_uri).group(2)

                    if (
                        "releaseVersion" in component["PipelineComponent"]
                        or "path=" in pipeline_id
                        or pipeline_id not in self.pipelines_build_path
                    ):
                        # Check if this shared pipelineComponent has configs
                        dependent_pipeline_regex_pattern = (
                            "(^[0-9]+?\/pipelines\/[-_.A-Za-z0-9 \/]+)$"
                            "|^.*projectSubscriptionProjectId=([0-9]+).*path=([-_.A-Za-z0-9 \/]+).*$"
                            "|^.*path=([-_.A-Za-z0-9 \/]+).*projectSubscriptionProjectId=([0-9]+).*$"
                        )
                        print(f"Parsing basepipeline: {pipeline_uri}")
                        search_regex_id = re.search(dependent_pipeline_regex_pattern, pipeline_uri)
                        if bool(search_regex_id) and "configPath" in component["PipelineComponent"]:
                            generate_pipeline_config_from_pipeline_component = True
                            shared_pipeline_id = ""
                            if bool(search_regex_id.group(1)):
                                shared_pipeline_id = search_regex_id.group(1)
                            elif bool(search_regex_id.group(2)) and bool(search_regex_id.group(3)):
                                shared_pipeline_id = f"{search_regex_id.group(2)}/{search_regex_id.group(3)}"
                            elif bool(search_regex_id.group(4)) and bool(search_regex_id.group(5)):
                                shared_pipeline_id = f"{search_regex_id.group(5)}/{search_regex_id.group(4)}"
                            else:
                                generate_pipeline_config_from_pipeline_component = False
                            self.pipeline_to_dbfs_config_path[shared_pipeline_id] = component["PipelineComponent"][
                                "configPath"
                            ]

                            print(f"Shared pipeline id for {pipeline_uri} is {shared_pipeline_id}")

                        print(f"    Pipeline {pipeline_id} might be shared, checking if it exists in DBFS")

                        # This jar is from a shared pipeline, check if jar exists in DBFS
                        try:
                            if self.dbfs_service.get_status(component["PipelineComponent"]["path"]):
                                print("    Dependent package exists on DBFS already, continuing with next pipeline")
                                continue
                        except HTTPError as e:
                            dependent_build_jar_found = False
                            if e.response.status_code == 404:
                                # Pipeline not found on DBFS. Check if it is present in dependency folder
                                # and try to build it ourselves
                                print(
                                    f"    Pipeline {pipeline_id} not found in DFBS, "
                                    f"searching in dependent project directory"
                                )
                                for project in self.dependent_projects.values():
                                    if (
                                        pipeline_id in project.pipelines
                                        and self.project_language == project.project_language
                                    ):
                                        print("    Building dependent project's pipeline:")
                                        (
                                            dependent_build_status,
                                            dependent_build_paths,
                                        ) = project.build(
                                            {
                                                k: v
                                                for (k, v) in project.pipelines.items()
                                                if pipeline_id in project.pipelines
                                            },
                                            False,
                                        )
                                        if dependent_build_status:
                                            dependent_build_jar_found = True
                                            self.dependent_pipelines_build_path[pipeline_id] = dependent_build_paths[
                                                pipeline_id
                                            ]
                            if not dependent_build_jar_found:
                                pipelines_upload_failures_job[pipeline_id].append(e.response.text)
                                pipelines_upload_failures[pipeline_id].append(e.response.text)
                        except Exception as ex:
                            self.pipelines_build_path[pipeline_id]["uploaded"] = False
                            pipelines_upload_failures_job[pipeline_id].append(str(ex))
                            pipelines_upload_failures[pipeline_id].append(str(ex))
                    if pipeline_id in self.pipelines_build_path or pipeline_id in self.dependent_pipelines_build_path:
                        pipelines_build_path = (
                            self.pipelines_build_path
                            if pipeline_id in self.pipelines_build_path
                            else self.dependent_pipelines_build_path
                        )
                        source_path = pipelines_build_path[pipeline_id]["source_absolute"]
                        target_path = component["PipelineComponent"]["path"]
                        if (
                            not pipelines_build_path[pipeline_id]["uploaded"]
                            or target_path not in self.uploaded_target_paths
                        ):
                            print(
                                "    Uploading %s to %s"
                                % (
                                    pipelines_build_path[pipeline_id]["source"],
                                    target_path,
                                )
                            )

                            try:
                                self.dbfs_service.put(target_path, overwrite=True, src_path=source_path)
                                pipelines_build_path[pipeline_id]["uploaded"] = True
                                self.uploaded_target_paths.add(target_path)
                            except HTTPError as e:
                                pipelines_upload_failures_job[pipeline_id].append(e.response.text)
                                pipelines_upload_failures[pipeline_id].append(e.response.text)
                            except Exception as ex:
                                pipelines_build_path[pipeline_id]["uploaded"] = False
                                pipelines_upload_failures_job[pipeline_id].append(str(ex))
                                pipelines_upload_failures[pipeline_id].append(str(ex))

            job_request = job_definition["request"]
            if "CreateNewJobRequest" in job_request.keys():
                job_request = job_definition["request"]["CreateNewJobRequest"]

            job_request["version"] = "2.1"
            # Upload configuration json files to dbfs

            print(f"Pipeline to DBFS config map: {self.pipeline_to_dbfs_config_path}")
            local_path_dbfs_map = dict()
            if generate_pipeline_config_from_pipeline_component:
                for (
                    base_pipeline_id,
                    local_config_path,
                ) in self.pipeline_to_local_config_path.items():
                    if base_pipeline_id in self.pipeline_to_dbfs_config_path:
                        local_path_dbfs_map[local_config_path] = self.pipeline_to_dbfs_config_path[base_pipeline_id]
            else:
                json_configs = self._get_spark_parameter_files(job_request["tasks"], ".json")

                local_path_dbfs_map = self._construct_local_config_to_dbfs_config_path(json_configs)

            for local_json_directory, dbfs_directory in local_path_dbfs_map.items():
                files = [
                    join(local_json_directory, f)
                    for f in listdir(local_json_directory)
                    if isfile(join(local_json_directory, f))
                ]
                for file_path in files:
                    dbfs_file_path = join(dbfs_directory, basename(file_path))
                    try:
                        print(f"    Uploading {file_path} to {dbfs_file_path}")
                        self.dbfs_service.put(dbfs_file_path, overwrite=True, src_path=file_path)
                    except Exception as e:
                        print(f"    Failed to upload {file_path} to {dbfs_file_path}. Exception: {e}")

            if len(pipelines_upload_failures_job) > 0:
                print("\n[bold red] Package upload failed or doesn't exist for one or more pipelines [/bold red]")
                for pipeline_id, errors in pipelines_upload_failures_job.items():
                    print("\n[bold red] %s: Exceptions: %s [/bold red]" % (pipeline_id, "\n".join(errors)))

                # Process next job
                continue

            limit = 25
            current_offset = 0
            found_job = None
            try:
                while found_job is None:
                    print(
                        f"    Querying existing jobs to find current job: Offset: {current_offset}, Pagesize: {limit}"
                    )
                    response = self.jobs_service.list_jobs(limit=limit, offset=current_offset, version="2.1")
                    current_offset += limit

                    found_jobs = response["jobs"] if "jobs" in response else []
                    for potential_found_job in found_jobs:
                        if potential_found_job["settings"]["name"] == job_request["name"]:
                            found_job = potential_found_job
                            break

                    if len(found_jobs) <= 0:
                        break

                job_request = job_definition["request"]
                if "CreateNewJobRequest" in job_request.keys():
                    job_request = job_definition["request"]["CreateNewJobRequest"]

                if found_job is None:
                    print("    Creating a new job: %s" % (job_request["name"]))
                    self.jobs_service.create_job(**job_request)
                else:
                    print("    Updating an existing job: %s" % (job_request["name"]))
                    self.jobs_service.reset_job(found_job["job_id"], new_settings=job_request, version="2.1")
            except Exception as e:
                print(
                    f"\n[bold red] Create/Update for job: {job_request['name']} failed with exception: {e} [/bold red]"
                )
                job_update_failures[job_request["name"]] = str(e)

        if len(pipelines_upload_failures) == 0 and len(job_update_failures) == 0:
            print("\n[bold blue][DONE]: Deployment completed successfully![/bold blue]")
        else:
            print("\n[bold red] Deployment failed![/bold red]")
            if len(pipelines_upload_failures) > 0:
                print("   Pipeline failures: %s" % (" ,".join(pipelines_upload_failures.keys())))
            if len(job_update_failures) > 0:
                print("   Create/Update failed for jobs: %s" % (" ,".join(job_update_failures.keys())))
            sys.exit(1)

    def test(self, build_jars: str = "", pipelines=None):
        if not pipelines:  # if pipelines not provided test all pipelines
            pipelines = self.pipelines
            pipelines_count = self.pipelines_count
            print("\n[INFO]: Testing All pipelines")
        else:
            pipeline_filter = [x.strip() for x in pipelines.split(",")]
            pipelines = {k: v for k, v in self.pipelines.items() if k.split("/")[1] in pipeline_filter}
            pipelines_count = len(pipelines)
            print(f"\n[INFO]: Pipeline Filters passed [{len(pipeline_filter)}]: {pipeline_filter}")
            print(f"[INFO]: Pipelines found [{pipelines_count}]: {list(pipelines.keys())}")

            if pipelines_count != len(pipeline_filter):
                print("\n[ERROR]: Filtered pipelines doesn't match with passed filter"
                      "\n Please check for any typo in Pipeline Filter")
                sys.exit(1)

        if self._verify_unit_test_env():
            unit_test_results = {}

            for pipeline_i, (path_pipeline, pipeline) in enumerate(pipelines.items()):
                print("\n  Unit Testing pipeline %s [%s/%s]" % (path_pipeline, pipeline_i + 1, pipelines_count))

                path_pipeline_absolute = os.path.join(os.path.join(self.path_root, path_pipeline), "code")
                if self.project_language == "python":
                    if os.path.isfile(os.path.join(path_pipeline_absolute, f"test{os.sep}TestSuite.py")):
                        # unique_key_for_jars = \
                        self._setJarsNeededForUT(build_jars)
                        unit_test_results[path_pipeline] = self.test_python(path_pipeline_absolute, path_pipeline)
                        self.removeJarsKeyFromEnv()
                elif self.project_language == "scala":
                    unit_test_results[path_pipeline] = self.test_scala(path_pipeline_absolute)
            is_any_ut_failed = False
            for path_pipeline, return_code in unit_test_results.items():
                if return_code not in (0, 5):
                    is_any_ut_failed = True
                    print(
                        f"\n[bold red] Unit test for pipeline: {path_pipeline} failed "
                        f"with return code {return_code}![/bold red]"
                    )
                else:
                    print(f"\n[bold blue] Unit test for pipeline: {path_pipeline} succeeded.[/bold blue]")

            if is_any_ut_failed:
                sys.exit(1)
        else:
            sys.exit(1)

    def get_python_dependencies(self, path_pipeline_absolute, path_pipeline):
        python_dependencies = []
        if (
            Process.process_sequential(
                [
                    # Export dependencies to egg_info/requires.txt
                    Process(
                        [self.python_cmd, "setup.py", "-q", "egg_info"],
                        path_pipeline_absolute,
                        is_shell=(self.operating_system == "win32"),
                        running_message="    Getting Python dependencies...",
                    )
                ]
            )
            == 0
        ):
            egg_info_requires_glob = glob(f"{path_pipeline_absolute}/**/*.egg-info/requires.txt", recursive=True)
            if not egg_info_requires_glob:
                print(f"Failed to get python dependencies for {path_pipeline}")
            else:
                egg_info_requires = egg_info_requires_glob[0]
                for line in open(egg_info_requires, "r"):
                    if not line.lstrip().startswith("[") and line.strip():
                        python_dependencies.append(line.strip())
        else:
            print(f"Failed to get python dependencies for {path_pipeline}")

        print(f"    Dependencies found for {path_pipeline}: {python_dependencies}")
        return python_dependencies

    def build_python(self, path_pipeline_absolute, path_pipeline):
        python_dependencies = self.get_python_dependencies(path_pipeline_absolute, path_pipeline)

        install_python_dependencies_cmd = (
            [self.pip_cmd, "install", "-q"] + python_dependencies
            if python_dependencies
            else [
                # this is just a fallback in case reading dependencies from setup.py fails
                self.pip_cmd,
                "install",
                "-q",
                "pyhocon",
                "prophecy-libs",
                "pytest",
                "pytest-html",
            ]
        )

        return Process.process_sequential(
            [
                # Extract the install_requires and extra_requires and install them
                Process(
                    install_python_dependencies_cmd,
                    path_pipeline_absolute,
                    is_shell=(self.operating_system == "win32"),
                ),
                # Check for compilation errors
                Process(
                    [self.python_cmd, "-m", "compileall", ".", "-q"],
                    path_pipeline_absolute,
                    is_shell=(self.operating_system == "win32"),
                ),
                # Generate wheel
                Process(
                    [self.python_cmd, "setup.py", "bdist_wheel"],
                    path_pipeline_absolute,
                    is_shell=(self.operating_system == "win32"),
                ),
            ]
        )

    def build_scala(self, path_pipeline_absolute):
        return Process.process_sequential(
            [
                Process(
                    ["mvn", "clean", "package", "-q", "-DskipTests"],
                    path_pipeline_absolute,
                    is_shell=(self.operating_system == "win32"),
                )
            ]
        )

    def test_scala(self, path_pipeline_absolute):
        return Process.process_sequential(
            [
                Process(
                    ["mvn", "test", "-q", "-Dfabric=" + self.fabric.strip()],
                    path_pipeline_absolute,
                    is_shell=(self.operating_system == "win32"),
                )
            ]
        )

    def test_python(self, path_pipeline_absolute, path_pipeline):
        return Process.process_sequential(
            [
                # Install dependencies of particular pipeline
                # 1. Export dependencies to egg_info/requires.txt
                Process(
                    [self.python_cmd, "setup.py", "-q", "egg_info"],
                    path_pipeline_absolute,
                    is_shell=(self.operating_system == "win32"),
                ),
                # 2. Extract the install_requires and extra_requires and install them
                Process(
                    [self.pip_cmd, "install", "-q"]
                    + self.get_python_dependencies(path_pipeline_absolute, path_pipeline),
                    path_pipeline_absolute,
                    is_shell=(self.operating_system == "win32"),
                ),
                # Run the unit test
                Process(
                    [
                        self.python_cmd,
                        "-m",
                        "pytest",
                        "-v",
                        f"test{os.sep}TestSuite.py",
                    ],
                    path_pipeline_absolute,
                    is_shell=(self.operating_system == "win32"),
                ),
            ]
        )

    def _verify_unit_test_env(self):
        self.fabric = os.environ.get("FABRIC_NAME")

        if self.fabric is None:
            self.fabric = "default"
            os.environ["FABRIC_NAME"] = self.fabric
        return True

    def _parse_project(self, ignore_parse_errors=False):
        self.pipelines: Dict = {}
        self.jobs: Dict = {}
        with open(self.path_project, "r") as _in:
            self.project = yaml.safe_load(_in)
            self.jobs = dict(
                (job_name, job)
                for job_name, job in self.project["jobs"].items()
                if "scheduler" in job and "Databricks" in job["scheduler"]
            )
            self.pipelines = self.project["pipelines"]
            self.project_language = self.project["language"]
            self.pipeline_configurations = dict(self.project.get("pipelineConfigurations", []))
            self.pipeline_to_local_config_path = {}
            self.pipeline_to_dbfs_config_path = {}
            for (
                pipeline_config_path,
                pipeline_config_object,
            ) in self.pipeline_configurations.items():
                self.pipeline_to_local_config_path[pipeline_config_object["basePipeline"]] = os.path.join(
                    self.path_root, pipeline_config_path
                )
        if self.project_language not in ("python", "scala"):
            self._error(f"Language: [i]{self.project_language}[/i] not supported by Prophecy-Build-Tool right now.")

        self.pipelines_count = len(self.pipelines)
        self.jobs_count = len(self.jobs)

        jobs_str = ", ".join(map(lambda job: job["name"], self.jobs.values()))
        print(f"Project name: {self.project['name']}")
        print("Found %s jobs: %s" % (self.jobs_count, jobs_str))

        pipelines_str = ", ".join(
            map(
                lambda pipeline: "%s (%s)" % (pipeline["name"], pipeline["language"]),
                self.pipelines.values(),
            )
        )
        print("Found %s pipelines: %s" % (self.pipelines_count, pipelines_str))
        self._verify_project_structure(ignore_parse_errors)

    def _verify_project_structure(self, ignore_parse_errors=False):
        for path_pipeline, pipeline in self.pipelines.items():
            path_pipeline_absolute = os.path.join(os.path.join(self.path_root, path_pipeline), "code")
            pipeline_dependencies_file = "setup.py" if self.project_language == "python" else "pom.xml"

            if not os.path.isfile(os.path.join(path_pipeline_absolute, pipeline_dependencies_file)):
                print(f"\n[bold red]Pipeline {path_pipeline} does not exist or is corrupted. [/bold red]")
                if not ignore_parse_errors:
                    sys.exit(1)
                else:
                    print(
                        f"\n[bold yellow] Ignoring Parse Error for {path_pipeline} as --ignore-parse-errors is passed ["
                        f"/bold yellow]"
                    )

        for path_job, job in self.jobs.items():
            path_job_definition = os.path.join(
                os.path.join(self.path_root, path_job),
                f"code{os.sep}databricks-job.json",
            )

            if not os.path.isfile(path_job_definition):
                print(f"\n[bold red]Job {path_job} does not exist or is corrupted. [/bold red]")
                if not ignore_parse_errors:
                    sys.exit(1)
                    print(
                        f"\n[bold yellow] Ignoring Parse Error for {path_job} as --ignore-parse-errors is passed ["
                        f"/bold yellow]"
                    )

    def _get_spark_parameter_files(self, tasks_list, file_extension):
        result = []
        package_field = "spark_jar_task" if self.project_language == "scala" else "python_wheel_task"
        for task in tasks_list:
            if package_field in task and "parameters" in task[package_field]:
                params = [
                    file.replace("/dbfs/", "dbfs:/")
                    for file in list(task[package_field]["parameters"])
                    if file.endswith(file_extension)
                ]
                result += params
        return result

    def _construct_local_config_to_dbfs_config_path(self, json_configs):
        local_config_to_dbfs_config_path_map = {}
        for json_config in json_configs:
            for (
                pipeline_id,
                pipeline_config_path,
            ) in self.pipeline_to_local_config_path.items():
                if pipeline_id in json_config:
                    local_config_to_dbfs_config_path_map[pipeline_config_path] = os.path.dirname(json_config)

        return local_config_to_dbfs_config_path_map

    @classmethod
    def _verify_databricks_configs(cls, exit_on_failure=True):
        host = os.environ.get("DATABRICKS_HOST")
        token = os.environ.get("DATABRICKS_TOKEN")

        if host is None or token is None:
            if exit_on_failure:
                cls._error(
                    "[i]DATABRICKS_HOST[/i] & [i]DATABRICKS_TOKEN[/i] environment variables are required to "
                    "deploy your Databricks Workflows"
                )
            else:
                return False
        return True

    def _verify_project(self):
        if not os.path.isfile(self.path_project):
            self._error(
                "Missing [i]pbt_project.yml[/i] file. Are you sure you pointed pbt into a Prophecy project? "
                "Current path [u]%s[/u]" % self.path_root
            )

    @classmethod
    def _error(cls, message: str):
        print("[bold red]ERROR[/bold red]:", message)
        sys.exit(1)

    def _setJarsNeededForUT(self, build_jars):
        # import random
        # uniqueKey = random.random()
        # jars_unique_key: str = f"driver_library_path_{uniqueKey}"
        os.environ["SPARK_JARS_CONFIG"] = build_jars if build_jars else ""
        # return jars_unique_key

    def removeJarsKeyFromEnv(self):
        del os.environ["SPARK_JARS_CONFIG"]



import json
from abc import ABC
from concurrent.futures import ThreadPoolExecutor
from typing import Dict, List, Optional

from requests import HTTPError

from ...client.rest_client_factory import RestClientFactory
from ...deployment import JobInfoAndOperation, OperationType, JobData, EntityIdToFabricId
from ...entities.project import Project
from ...utility import custom_print as log, Either, is_online_mode
from ...utils.constants import COMPONENTS_LITERAL
from ...utils.exceptions import InvalidFabricException
from ...utils.project_config import JobInfo, ProjectConfig, await_futures_and_update_states
from ...utils.project_models import DbtComponentsModel, ScriptComponentsModel, StepMetadata, Operation, StepType, \
    Colors

SCRIPT_COMPONENT = "ScriptComponent"
COMPONENTS = "components"
FABRIC_ID = "fabric_id"
DBT_COMPONENT = "DBTComponent"


def get_fabric_label(name: str, id: str):
    if name is not None and len(name) > 0:
        return name
    else:
        return id


# todo add a new abstract class for DatabricksJobs and AirflowJobs
class DatabricksJobs(JobData, ABC):
    def __init__(self, pbt_job_json: Dict[str, str], databricks_job: str, fabric_override: Optional[str] = None):
        self.pbt_job_json = pbt_job_json
        self.databricks_job = databricks_job
        self.fabric_override = fabric_override

        try:
            self.databricks_job_json = json.loads(databricks_job)
        except Exception:
            self.databricks_job_json = {}

    @property
    def is_valid_job(self):
        return len(self.databricks_job_json) > 0 and len(self.pbt_job_json) > 0

    @property
    def fabric_id(self):
        if self.fabric_override is not None and len(self.fabric_override) > 0:
            return self.fabric_override
        else:
            fabric_id = self.pbt_job_json.get('fabricUID', None)

            if fabric_id is not None:
                return str(fabric_id)

        return fabric_id

    def dag_name(self) -> Optional[str]:
        pass

    def has_dbt_component(self) -> bool:
        pass

    def is_disabled(self) -> bool:
        pass

    def job_files(self):
        pass

    def is_enabled(self) -> bool:
        pass

    def validate_prophecy_managed_checksum(self, salt: str) -> bool:
        pass

    @property
    def pipeline_and_fabric_ids(self) -> List[EntityIdToFabricId]:
        pipeline_and_fabric_ids = []

        for pipeline_id in self.pipelines:
            pipeline_and_fabric_ids.append(EntityIdToFabricId(pipeline_id, self.fabric_id))

        return pipeline_and_fabric_ids

    @property
    def databricks_json(self):
        return self.databricks_job_json.get('request', None)

    @property
    def acl(self):
        return self.databricks_job_json.get('request', {}).get('accessControlList', None)

    @property
    def get_secret_scope(self):
        return self.databricks_job_json.get('secret_scope', None)

    @property
    def pipelines(self):
        return list(set(self.pbt_job_json.get('pipelines', None)))

    @property
    def name(self):
        return self.pbt_job_json.get('name', None)

    @property
    def is_paused(self):
        return self.pbt_job_json.get('enabled', False)


class DatabricksJobsDeployment:
    _ADD_JOBS_STEP_ID = "add-db-jobs"
    _REFRESH_JOBS_STEP_ID = "refresh-db-jobs"
    _DELETE_JOBS_STEP_ID = "delete-db-jobs"
    _PAUSE_JOBS_STEP_ID = "pause-db-jobs"

    def __init__(self, project: Project, project_config: ProjectConfig):
        self.project = project

        self.project_config = project_config
        self.deployment_state = project_config.jobs_state
        self.deployment_run_override_config = project_config.configs_override
        self.fabric_configs = project_config.fabric_config

        self._pipeline_configurations = self.project.pipeline_configurations
        self._rest_client_factory = RestClientFactory.get_instance(RestClientFactory,
                                                                   fabric_config=self.project_config.fabric_config)

        (self._db_jobs, self._skipping_jobs) = self._initialize_db_jobs()
        self.valid_databricks_jobs, self._databricks_jobs_without_code = self._initialize_valid_databricks_jobs()

    def summary(self):

        summary = []

        summary.extend([f"Adding job {id}" for id in list(self._add_jobs().keys())])
        summary.extend([f"Refreshing job {id}" for id in list(self._refresh_jobs().keys())])
        summary.extend([f"Deleting job {id}" for id in list(self._delete_jobs().keys())])
        summary.extend([f"Pausing job {job.id}" for job in list(self._pause_jobs())])

        return summary

    @property
    def _operation_to_step_id(self):
        return {
            Operation.Add: self._ADD_JOBS_STEP_ID,
            Operation.Refresh: self._REFRESH_JOBS_STEP_ID,
            Operation.Remove: self._DELETE_JOBS_STEP_ID,
            Operation.Pause: self._PAUSE_JOBS_STEP_ID
        }

    def headers(self) -> List[StepMetadata]:
        job_types = {
            Operation.Add: self._add_jobs(),
            Operation.Refresh: self._refresh_jobs(),
            Operation.Remove: self._delete_jobs(),
            Operation.Pause: self._pause_jobs()
        }

        all_headers = []

        for job_action, jobs in job_types.items():
            if jobs:
                len_jobs = len(jobs)
                job_header_suffix = f'{len_jobs} Databricks job' if len_jobs == 1 else f'{len_jobs} Databricks jobs'

                step_id = self._operation_to_step_id[job_action]

                all_headers.append(
                    StepMetadata(step_id, f"{job_action.value} {job_header_suffix}",
                                 job_action, StepType.Job)
                )

        return all_headers

    def deploy(self) -> List[Either]:
        if len(self.headers()) > 0:
            log(f"{Colors.OKBLUE}\nDeploying databricks jobs{Colors.ENDC}\n\n")

        responses = self._deploy_add_jobs() + self._deploy_refresh_jobs() + \
            self._deploy_delete_jobs() + self._deploy_pause_jobs() + self._deploy_skipping_jobs()

        return responses

    def get_databricks_client(self, fabric_id):
        return self._rest_client_factory.databricks_client(fabric_id)

    @property
    def databricks_job_json_for_refresh_and_new_jobs(self) -> Dict[str, DatabricksJobs]:
        return self._databricks_jobs_data(self._add_and_refresh_job_ids())

    def _deploy_skipping_jobs(self) -> List[Either]:
        responses = []
        allowed_fabrics = self.fabric_configs.list_all_fabrics()

        if len(self._skipping_jobs) > 0:
            log(f"{Colors.OKBLUE}\n\nSkipping databricks jobs{Colors.ENDC}\n")

        for job_id, job in self._skipping_jobs.items():
            log(f"[SKIP] Job {job_id} skipped as it belongs to fabric-id {job.fabric_id} but allowed fabric-ids are {allowed_fabrics}")

        return responses

    def _initialize_db_jobs(self):
        jobs = {}
        skipping_jobs = {}
        for job_id, pbt_job_json in self.project.jobs.items():

            fabric_override = self.deployment_run_override_config.find_fabric_override_for_job(job_id)
            fabric_override = str(fabric_override) if fabric_override is not None else None

            job_fabric = pbt_job_json.get('fabricUID', None)
            job_fabric = str(job_fabric) if job_fabric is not None else None

            does_fabric_exist = self.fabric_configs.get_fabric(
                job_fabric) is not None or self.fabric_configs.get_fabric(fabric_override) is not None

            databricks_job = self.project.load_databricks_job(job_id)

            if 'Databricks' in pbt_job_json.get('scheduler', None) and databricks_job is not None:
                if self.deployment_run_override_config.is_job_to_run(job_id):

                    if does_fabric_exist:
                        jobs[job_id] = DatabricksJobs(pbt_job_json, databricks_job, fabric_override)

                    else:
                        skipping_jobs[job_id] = DatabricksJobs(pbt_job_json, databricks_job, fabric_override)

        return jobs, skipping_jobs

    def _initialize_valid_databricks_jobs(self):

        valid_databricks_jobs = {}
        databricks_jobs_without_code = {}

        for job_id, job_data in self._db_jobs.items():
            if job_data.is_valid_job:
                valid_databricks_jobs[job_id] = job_data

            else:
                databricks_jobs_without_code[job_id] = job_data

        return valid_databricks_jobs, databricks_jobs_without_code

    def _add_and_refresh_job_ids(self) -> List[str]:
        return list(set(list(self._add_jobs().keys()) + list(self._refresh_jobs().keys())))

    def _databricks_jobs_data(self, jobs: List[str]) -> Dict[str, DatabricksJobs]:

        return {
            job: self.valid_databricks_jobs.get(job, {}) for job in jobs
        }

    def _refresh_jobs(self) -> Dict[str, JobInfo]:
        return {
            job_info.id: job_info for job_info in self.deployment_state.databricks_jobs
            if any(
                job_info.id == job_id and job_info.fabric_id == job_data.fabric_id
                for job_id, job_data in self.valid_databricks_jobs.items()
            )
        }

    def _pause_jobs(self) -> List[JobInfo]:

        return [
            databricks_job for databricks_job in self.deployment_state.databricks_jobs
            if any(
                databricks_job.id == job_id and databricks_job.fabric_id != job_data.fabric_id
                for job_id, job_data in self.valid_databricks_jobs.items()
            )
        ]

    def _add_jobs(self) -> Dict[str, DatabricksJobs]:

        return {
            job_id: job_data for job_id, job_data in self.valid_databricks_jobs.items()
            if self.deployment_state.contains_job(job_id, str(job_data.fabric_id)) is False
        }

    def _delete_jobs(self) -> Dict[str, JobInfo]:
        all_jobs = {**self.valid_databricks_jobs, **self._databricks_jobs_without_code}
        return {
            job.id: job for job in self.deployment_state.databricks_jobs
            if not any(job.id == job_id for job_id in
                       all_jobs.keys())
        }

    '''Deploy Jobs '''

    def _deploy_add_job(self, job_id, job_data, step_id):
        fabric_id = job_data.fabric_id
        fabric_config = self.project_config.fabric_config.get_fabric(fabric_id)
        fabric_name = fabric_config.name if fabric_config is not None else None

        fabric_label = get_fabric_label(fabric_name, fabric_id)

        try:
            client = self.get_databricks_client(fabric_id)
            operation = OperationType.CREATED

            scheduler_job_id = None
            if not is_online_mode() and not self.project_config.based_on_file:
                scheduler_job_id = client.find_job(job_data.name)
                if scheduler_job_id:
                    client.reset_job(scheduler_job_id, job_data.databricks_json)
                    log(f"{Colors.OKGREEN}Refreshed job {job_id} in fabric {fabric_label} databricks job-id:{scheduler_job_id}{Colors.ENDC}",
                        step_id=step_id)
                    operation = OperationType.REFRESH

            if not scheduler_job_id:
                response = client.create_job(job_data.databricks_json)
                log(f"{Colors.OKGREEN}Created job {job_id} in fabric {fabric_label} response {response['job_id']}{Colors.ENDC}",
                    step_id=step_id)
                scheduler_job_id = response['job_id']

            job_info = JobInfo.create_job(job_data.name, job_id, fabric_id, scheduler_job_id,
                                          self.project.release_tag, job_data.is_paused)

            return Either(right=JobInfoAndOperation(job_info, operation))

        except Exception as e:

            log(f"{Colors.FAIL}Error creating job {job_id} in fabric {fabric_id}{Colors.ENDC}", exception=e,
                step_id=step_id)
            return Either(left=e)

    def _deploy_add_jobs(self):

        with ThreadPoolExecutor(max_workers=3) as executor:
            futures = []

            for job_id, job_data in self._add_jobs().items():
                fabric_id = job_data.fabric_id

                if fabric_id is not None:
                    futures.append(executor.submit(
                        lambda j_id=job_id, j_data=job_data: self._deploy_add_job(j_id, j_data,
                                                                                  self._ADD_JOBS_STEP_ID)))
                else:
                    log(f"{Colors.WARNING} Invalid fabric {fabric_id}, skipping job creation for job_id {job_id}{Colors.ENDC}",
                        step_id=self._ADD_JOBS_STEP_ID)

        return await_futures_and_update_states(futures, self._operation_to_step_id[Operation.Add])

    def _deploy_refresh_jobs(self):
        futures = []

        with ThreadPoolExecutor(max_workers=3) as executor:

            for job_id, job_info in self._refresh_jobs().items():
                fabric_id = job_info.fabric_id

                if fabric_id is not None:
                    futures.append(executor.submit(
                        lambda j_id=job_id, j_info=job_info: self._reset_and_patch_job(j_id, j_info)))
                else:
                    log(f"${Colors.OKCYAN}Invalid fabric {fabric_id}, skipping job refresh for job_id {job_id}{Colors.ENDC}")

        return await_futures_and_update_states(futures, self._operation_to_step_id[Operation.Refresh])

    def _reset_and_patch_job(self, job_id, job_info):
        fabric_id = job_info.fabric_id
        fabric_config = self.fabric_configs.get_fabric(fabric_id)
        fabric_name = fabric_config.name if fabric_config is not None else None
        fabric_label = get_fabric_label(fabric_name, fabric_id)

        def log_error(message, exc, step_id=self._REFRESH_JOBS_STEP_ID):
            log(message, exception=exc, step_id=step_id)

        def log_success(message):
            log(message, step_id=self._REFRESH_JOBS_STEP_ID)

        try:
            client = self.get_databricks_client(fabric_id)
            job_data = self.valid_databricks_jobs.get(job_id)
            client.reset_job(job_info.external_job_id, job_data.databricks_json)

            if job_data.acl:
                log_success(
                    f"{Colors.OKGREEN}Refreshed job {job_id} with external job id {job_info.external_job_id} patching job acl{Colors.ENDC}")
                try:
                    client.patch_job_acl(job_info.external_job_id, job_data.acl)
                except Exception as e:
                    log_error(
                        f"{Colors.FAIL}Error patching job acl for job {job_id} in fabric {fabric_label}{Colors.ENDC}",
                        e)
                    return Either(left=e)
            else:
                log_success(
                    f"{Colors.OKGREEN}Refreshed job {job_id} with external job id {job_info.external_job_id}, no acl found{Colors.ENDC}")

            return Either(right=JobInfoAndOperation(job_info, OperationType.REFRESH))

        except HTTPError as e:
            response = e.response.content.decode('utf-8')

            if e.response.status_code == 400 and f"Job {job_info.external_job_id} does not exist." in response:
                log_success(
                    f"{Colors.OKGREEN}Job {job_id} external_job_id {job_info.external_job_id} deleted in fabric {fabric_label}, trying to recreate it{Colors.ENDC}")
                return self._deploy_add_job(job_id, self.valid_databricks_jobs.get(job_id),
                                            step_id=self._REFRESH_JOBS_STEP_ID)
            else:
                log_error(
                    f'{Colors.FAIL}Error while refreshing job with scheduler id {job_info.external_job_id} and id {job_info.id} for fabric {fabric_label}{Colors.ENDC}',
                    e)
                return Either(left=e)

        except Exception as e:
            log_error(
                f'{Colors.FAIL}Error while resetting job with {job_info.external_job_id} and job-id {job_id} for fabric {fabric_label}{Colors.ENDC}',
                e)
            return Either(left=e)

    def _deploy_delete_jobs(self):
        futures = []

        with ThreadPoolExecutor(max_workers=3) as executor:
            for job_id, job_info in self._delete_jobs().items():
                futures.append(executor.submit(lambda j_info=job_info: self._delete_job(j_info)))

        return await_futures_and_update_states(futures, self._operation_to_step_id[Operation.Remove])

    def _delete_job(self, job_info: JobInfo):
        fabric = job_info.fabric_id

        if fabric is None:
            msg = f"{Colors.WARNING}Invalid fabric {fabric} not deleting job {job_info.id} and databricks id {job_info.external_job_id}{Colors.ENDC}"
            log(msg, step_id=self._DELETE_JOBS_STEP_ID)
            return Either(left=InvalidFabricException(msg))

        def log_error(_msg, exc):
            log(_msg, exception=exc, step_id=self._DELETE_JOBS_STEP_ID)

        def log_success(message):
            log(message, step_id=self._DELETE_JOBS_STEP_ID)

        try:
            client = self.get_databricks_client(fabric)
            client.delete_job(job_info.external_job_id)
            log_success(
                f'{Colors.OKGREEN}Successfully deleted job with scheduler id {job_info.external_job_id} and id {job_info.id} for fabric {fabric}{Colors.ENDC}')
            return Either(right=JobInfoAndOperation(job_info, OperationType.DELETED))

        except HTTPError as e:

            response = e.response.content.decode('utf-8')

            if e.response.status_code == 400 and "does not exist." in response:
                log_success(
                    f"{Colors.OKGREEN}Job:{job_info.id} with external-id:{job_info.external_job_id}  already deleted, skipping{Colors.ENDC}")
                return Either(right=JobInfoAndOperation(job_info, OperationType.DELETED))
            log_error(
                f'{Colors.WARNING}Error on deleting job with scheduler id {job_info.external_job_id} and id {job_info.id} for fabric {fabric}{Colors.ENDC}',
                e)

            return Either(right=JobInfoAndOperation(job_info, OperationType.DELETED))

        except Exception as e:
            log_error(
                f'{Colors.FAIL}Error while deleting job with scheduler id {job_info.external_job_id} and id {job_info.id} for fabric {fabric}{Colors.ENDC}',
                e)
            return Either(left=e)

    def _deploy_pause_jobs(self):
        futures = []

        def pause_job(fabric_id, job_info):
            external_job_id = job_info.external_job_id
            fabric_name = get_fabric_label(self.fabric_configs.get_fabric(fabric_id).name, fabric_id)

            try:
                client = self.get_databricks_client(fabric_id)
                client.pause_job(external_job_id)
                log(f"{Colors.OKGREEN}Paused job {external_job_id} in fabric {fabric_name}{Colors.ENDC}",
                    step_id=self._PAUSE_JOBS_STEP_ID)
                return Either(right=JobInfoAndOperation(job_info, OperationType.REFRESH))

            except Exception as e:
                log(f"{Colors.WARNING}Error pausing job {external_job_id} in fabric {fabric_id}, Ignoring this {Colors.ENDC}",
                    exception=e,
                    step_id=self._PAUSE_JOBS_STEP_ID)
                return Either(right=JobInfoAndOperation(job_info, OperationType.REFRESH))

        with ThreadPoolExecutor(max_workers=3) as executor:

            for jobs_info in self._pause_jobs():
                fabric = jobs_info.fabric_id

                if fabric is not None:
                    futures.append(executor.submit(lambda f=fabric, j_info=jobs_info: pause_job(f, j_info)))
                else:
                    log(f"{Colors.WARNING}Invalid fabric {fabric} for job_id {jobs_info.id} {Colors.ENDC}",
                        step_id=self._PAUSE_JOBS_STEP_ID)

        return await_futures_and_update_states(futures, self._operation_to_step_id[Operation.Pause])


class DBTComponents:
    _DBT_SECRETS_COMPONENT_STEP_NAME = "DBTSecretsComponents"
    _DBT_PROFILES_COMPONENT_STEP_NAME = "DBTProfileComponents"

    def __init__(self, project: Project, databricks_jobs: DatabricksJobsDeployment,
                 project_config: ProjectConfig):
        self.databricks_jobs = databricks_jobs
        self.project = project
        self.project_config = project_config
        self.jobs_state = project_config.jobs_state
        self.fabrics_config = project_config.fabric_config

    def summary(self):
        summary = []

        for job_id, dbt_component_model in self.dbt_component_from_jobs.items():

            for components in dbt_component_model.components:
                if components.get('sqlFabricId', None) is not None and components.get('secretKey', None) is not None:
                    summary.append(
                        f"Uploading dbt secrets for job {job_id} component {components.get('nodeName')} ", )

                if components.get('profilePath', None) is not None and components.get('profileContent',
                                                                                      None) is not None:
                    summary.append(f"Uploading dbt profiles for job {job_id} component {components.get('nodeName')} ")

        return summary

    def headers(self):
        return self._dbt_secrets_headers() + self._dbt_profiles_to_build_headers()

    def deploy(self):
        return self._upload_dbt_secrets() + self._upload_dbt_profiles()

    @property
    def dbt_component_from_jobs(self) -> Dict[str, DbtComponentsModel]:
        job_id_to_dbt_components = {}

        for job_id, job_data in self.databricks_jobs.databricks_job_json_for_refresh_and_new_jobs.items():
            component_list = []

            for component in job_data.databricks_job_json.get(COMPONENTS_LITERAL, []):
                if DBT_COMPONENT in component:
                    component_list.append(component[DBT_COMPONENT])

            if len(component_list) > 0:
                # add prophecy_jobs as the default scope.
                secret_scope = job_data.get_secret_scope
                job_id_to_dbt_components[job_id] = DbtComponentsModel(job_data.fabric_id, secret_scope,
                                                                      component_list)

        return job_id_to_dbt_components

    def _upload_dbt_profiles(self):
        futures = []

        with ThreadPoolExecutor(max_workers=3) as executor:
            if len(self._dbt_profiles_to_build_headers()) > 0:
                log(f"\n\n{Colors.OKBLUE}Uploading DBT profiles{Colors.ENDC}\n\n")

            for job_id, dbt_component_model in self.dbt_component_from_jobs.items():
                for components in dbt_component_model.components:

                    if components.get('profilePath', None) is not None and components.get('profileContent',
                                                                                          None) is not None:
                        futures.append(executor.submit(
                            lambda model=dbt_component_model, comp=components: self._upload_dbt_profile(model, comp)))

        return await_futures_and_update_states(futures, self._DBT_PROFILES_COMPONENT_STEP_NAME)

    def _upload_dbt_profile(self, dbt_component_model: DbtComponentsModel, components: Dict[str, str]):
        try:
            client = self.databricks_jobs.get_databricks_client(dbt_component_model.fabric_id)
            client.upload_content(components['profileContent'], components['profilePath'])

            log(f"{Colors.OKGREEN}Successfully uploaded dbt profile {components['profilePath']}{Colors.ENDC}",
                step_id=self._DBT_PROFILES_COMPONENT_STEP_NAME)
            return Either(right=True)
        except Exception as e:
            log(f"{Colors.FAIL}Error while uploading dbt profile {components['profilePath']}{Colors.ENDC}", exception=e,
                step_id=self._DBT_PROFILES_COMPONENT_STEP_NAME)
            return Either(left=e)

    def _upload_dbt_secrets(self):
        futures = []

        with ThreadPoolExecutor(max_workers=3) as executor:
            if len(self._dbt_secrets_headers()) > 0:
                log(f"\n\n{Colors.OKBLUE}Uploading DBT secrets{Colors.ENDC}\n\n")

            for job_id, dbt_component_model in self.dbt_component_from_jobs.items():
                for dbt_component in dbt_component_model.components:
                    if dbt_component.get('sqlFabricId', None) is not None and dbt_component.get('secretKey',
                                                                                                None) is not None:
                        futures.append(
                            executor.submit(
                                lambda model=dbt_component_model, component=dbt_component: self._upload_dbt_secret(
                                    model, component)))

        return await_futures_and_update_states(futures, self._DBT_SECRETS_COMPONENT_STEP_NAME)

    def _upload_dbt_secret(self, dbt_component_model, dbt_component):
        try:
            self.databricks_jobs.get_databricks_client(str(dbt_component_model.fabric_id)) \
                .create_secret_scope_if_not_exist(dbt_component_model.secret_scope)

            sql_client = self.databricks_jobs.get_databricks_client(dbt_component['sqlFabricId'])
            git_token = self.fabrics_config.git_token_for_project(dbt_component['projectId'])
            master_token = f"{sql_client.token};{git_token}"

            sql_client.create_secret(dbt_component_model.secret_scope, dbt_component['secretKey'], master_token)

            log(f"{Colors.OKGREEN}Successfully uploaded dbt secret for component {dbt_component['nodeName']}{Colors.ENDC}",
                step_id=self._DBT_SECRETS_COMPONENT_STEP_NAME)

            return Either(right=True)

        except Exception as e:

            log(f"{Colors.FAIL}Error on uploading dbt secret for component {dbt_component['nodeName']} {Colors.ENDC}",
                exception=e,
                step_id=self._DBT_SECRETS_COMPONENT_STEP_NAME)

            return Either(left=e)

    def _dbt_profiles_to_build_headers(self):
        total_dbt_profiles = 0
        for job_id, dbt_component_model in self.dbt_component_from_jobs.items():

            for components in dbt_component_model.components:
                if components.get('profilePath', None) is not None and components.get('profileContent',
                                                                                      None) is not None:
                    total_dbt_profiles = total_dbt_profiles + 1

        if total_dbt_profiles > 0:
            return [StepMetadata(self._DBT_PROFILES_COMPONENT_STEP_NAME, f"Upload {total_dbt_profiles} dbt profiles",
                                 Operation.Upload,
                                 StepType.DbtProfile)]
        else:
            return []

    def _dbt_secrets_headers(self):
        total_dbt_secrets = 0
        for job_id, dbt_component_model in self.dbt_component_from_jobs.items():

            for components in dbt_component_model.components:
                if components.get('sqlFabricId', None) is not None and components.get('secretKey', None) is not None:
                    total_dbt_secrets = total_dbt_secrets + 1

        if total_dbt_secrets > 0:
            return [
                StepMetadata(self._DBT_SECRETS_COMPONENT_STEP_NAME, f"Upload {total_dbt_secrets} dbt secrets",
                             Operation.Upload,
                             StepType.DbtSecret)]
        else:
            return []


class ScriptComponents:
    _STEP_ID = "script_components"

    def __init__(self, project: Project, databricks_jobs: DatabricksJobsDeployment,
                 project_config: ProjectConfig):
        self.databricks_jobs = databricks_jobs
        self.project = project
        self.script_jobs = self._script_components_from_jobs()
        self.fabric_config = project_config.fabric_config
        self.deployment_state = project_config.jobs_state

    def summary(self) -> List[str]:
        script_summary = []

        for job_id, script_job_list in self.script_jobs.items():
            for script_job in script_job_list.scripts:
                script_summary.append(f"Uploading script {script_job['nodeName']} for job {job_id}")

        return script_summary

    def headers(self) -> List[StepMetadata]:
        all_scripts = sum(len(v.scripts) for v in self.script_jobs.values())

        if all_scripts > 0:
            return [StepMetadata(self._STEP_ID, f"Upload {all_scripts} script components", Operation.Upload,
                                 StepType.Script)]
        else:
            return []

    def deploy(self):
        futures = []

        if len(self.headers()) > 0:
            log(f"\n\n{Colors.OKBLUE}Uploading script components from job{Colors.ENDC}\n\n")

        with ThreadPoolExecutor(max_workers=10) as executor:
            for job_id, script_components in self._script_components_from_jobs().items():

                for scripts in script_components.scripts:
                    futures.append(executor.submit(
                        lambda s=scripts, fabric_id=script_components.fabric_id, j_id=job_id: self._upload_content(s,
                                                                                                                   fabric_id,
                                                                                                                   j_id)))

        return await_futures_and_update_states(futures, self._STEP_ID)

    def _upload_content(self, script: dict, fabric_id: str, job_id: str):
        node_name = script.get('nodeName')
        fabric_config = self.fabric_config.get_fabric(fabric_id)
        fabric_name = fabric_config.name if fabric_config is not None else None
        fabric_label = get_fabric_label(fabric_name, fabric_id)

        try:
            client = self.databricks_jobs.get_databricks_client(str(fabric_id))
            client.upload_content(content=script.get('content', None), path=script.get('path', None))

            log(f"{Colors.OKGREEN}Uploaded script component `{node_name}` to fabric `{fabric_label}` for job `{job_id}` to path `{script.get('path', None)}` {Colors.ENDC}",
                step_id=self._STEP_ID)
            return Either(right=True)
        except Exception as e:

            log(f"{Colors.FAIL}Error uploading script component `{node_name}` to fabric `{fabric_label}` {Colors.ENDC}",
                step_id=self._STEP_ID,
                exception=e)
            return Either(left=e)

    def _script_components_from_jobs(self) -> Dict[str, ScriptComponentsModel]:
        job_id_to_script_components = {}

        for jobs_id, job_data in self.databricks_jobs.databricks_job_json_for_refresh_and_new_jobs.items():
            script_component_list = []

            for components in job_data.databricks_job_json.get(COMPONENTS_LITERAL, []):
                if SCRIPT_COMPONENT in components:
                    script_component_list.append(components[SCRIPT_COMPONENT])

            fabric_id = job_data.fabric_id
            job_id_to_script_components[jobs_id] = ScriptComponentsModel(fabric_id, script_component_list)

        return job_id_to_script_components


class PipelineConfigurations:
    _STEP_ID = "pipeline_configurations"

    def __init__(self, project: Project, databricks_jobs: DatabricksJobsDeployment,
                 project_config: ProjectConfig):
        self.databricks_jobs = databricks_jobs
        self.pipeline_configurations = project.pipeline_configurations
        self.project_config = project_config
        self.project = project

    def summary(self) -> List[str]:
        summary = []
        for ids in self.pipeline_configurations.keys():
            summary.append(f"Uploading pipeline configurations for pipeline {ids}")

        return summary

    def headers(self) -> List[StepMetadata]:
        all_configs = [value for sublist in self.pipeline_configurations.values() for value in sublist]
        if len(all_configs) > 0:
            return [StepMetadata(self._STEP_ID,
                                 f"Upload {len(all_configs)} pipeline configurations",
                                 Operation.Upload, StepType.PipelineConfiguration)]
        else:
            return []

    def deploy(self):

        futures = []

        with ThreadPoolExecutor(max_workers=10) as executor:
            if len(self.pipeline_configurations.items()) > 0:
                count = sum(len(v) for v in self.pipeline_configurations.values())
                log(f"\n\n{Colors.OKBLUE} Uploading {count} pipeline configurations {Colors.ENDC}\n\n")

            def execute_job(_fabric_id, config_content, config_path):
                futures.append(executor.submit(
                    lambda f_id=str(_fabric_id), conf_content=config_content, conf_path=config_path:
                    self._upload_configuration(f_id, conf_content, conf_path)))

            for pipeline_id, configurations in self.pipeline_configurations.items():

                path = self.project_config.system_config.get_dbfs_base_path()
                pipeline_path = f'{path}/{self.project.project_id}/{self.project.release_version}/configurations/{pipeline_id}'

                for configuration_name, configuration_content in configurations.items():
                    configuration_path = f'{pipeline_path}/{configuration_name}.json'

                    for fabric_id in self.project_config.fabric_config.db_fabrics():
                        execute_job(fabric_id, configuration_content, configuration_path)

        return await_futures_and_update_states(futures, self._STEP_ID)

    def _upload_configuration(self, fabric_id, configuration_content, configuration_path):
        try:
            client = self.databricks_jobs.get_databricks_client(str(fabric_id))
            client.upload_content(configuration_content, configuration_path)

            log(f"{Colors.OKGREEN}Uploaded pipeline configuration on path {configuration_path}{Colors.ENDC}",
                step_id=self._STEP_ID)

            return Either(right=True)

        except Exception as e:
            log(f"{Colors.FAIL}Failed to upload pipeline configuration for path {configuration_path}{Colors.ENDC}",
                exception=e,
                step_id=self._STEP_ID)
            return Either(left=e)



import subprocess
import time


class Process:
    def __init__(
        self,
        process_args,
        current_working_directory,
        std_output=subprocess.PIPE,
        std_err=subprocess.PIPE,
        is_shell=False,
        running_message="",
    ):
        self.process_args = process_args
        self.current_working_directory = current_working_directory
        self.std_output = std_output
        self.std_err = std_err
        self.is_shell = is_shell
        self.running_message = running_message

    @staticmethod
    def process_sequential(processes, time_between_each_cmd=1):
        return_code = 0
        for process in processes:
            if process.running_message:
                print(process.running_message)
            result = subprocess.run(
                process.process_args,
                stdout=process.std_output,
                stderr=process.std_err,
                shell=process.is_shell,
                cwd=process.current_working_directory,
            )
            return_code, stdout, stderr = (
                result.returncode,
                result.stdout,
                result.stderr,
            )

            if stdout is not None:
                if stderr is not None and len(stderr) > 0:
                    print(
                        "   ",
                        "\n    ".join([line.decode("utf-8") for line in stdout.splitlines()]),
                    )
                    print(
                        "   ",
                        "\n    ".join([line.decode("utf-8") for line in stderr.splitlines()]),
                    )
                else:
                    print(
                        "   ",
                        "\n    ".join([line.decode("utf-8") for line in stdout.splitlines()]),
                    )

            if return_code != 0:
                break
            time.sleep(time_between_each_cmd)
        return return_code



import setuptools

with open("README.md", "r", encoding="utf-8") as _in:
    long_description = _in.read()

setuptools.setup(
    name="prophecy-build-tool",
    version="1.2.7-dev8",
    author="Prophecy",
    author_email="maciej@prophecy.io",
    description="Prophecy-build-tool (PBT) provides utilities to build and distribute projects created from the "
                "Prophecy IDE.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SimpleDataLabsInc/prophecy-build-tool",
    project_urls={
        "Bug Tracker": "https://github.com/SimpleDataLabsInc/prophecy-build-tool/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    install_requires=[
        "requests>=2.28.0",
        "PyYAML>=6.0",
        "databricks-cli>=0.17.7",
        "rich>=12.5.1",
        "wheel",
        "build",
        "google-cloud-secret-manager==2.16.2",
        "google-cloud-storage==2.10.0",
        "pydantic~=1.10",
        "pydantic-yaml==1.1.1",
        "boto3~=1.26.157",
        "tenacity==8.2.3",
    ],
    extras_require={
        "test": [
            "pytest-html",
            "pytest",
            "pyspark==3.3.0",
            "mock",  # or any other testing-specific packages you need
        ]
    },
    python_requires=">=3.7",
    entry_points="""
        [console_scripts]
        pbt=pbt:main
    """,
)



import os
import tempfile

import requests

from requests.auth import HTTPBasicAuth

from ..utils.exceptions import ArtifactDownloadFailedException


class NexusClient:

    def __init__(self, base_url: str, username: str, password: str, repository: str):

        self._base_url = base_url  # base-url https://execution.dev.cloud.prophecy.io/artifactory/
        self._username = username
        self._password = password
        self._repository = repository

    @classmethod
    def initialize_nexus_client(cls, project_config):
        nexus_config = project_config.system_config.nexus

        if nexus_config is not None:
            return NexusClient(nexus_config.url,
                               nexus_config.username,
                               nexus_config.password,
                               nexus_config.repository)
        else:
            raise Exception("Nexus config not found in deployment config, fallbacking to default settings.")

    def upload_file(self, file_path: str, project_id: str, pipeline_id: str, release_version: str, file_name: str):
        headers = {
            'accept': 'application/json'
        }
        url = f'{self._base_url}/service/rest/v1/components?repository={self._repository}'
        data = {
            'raw.directory': f'/prophecy/artifact/{self._create_file_path(project_id, release_version, pipeline_id)}',
            'raw.asset1.filename': file_name,
        }

        files = {
            'raw.asset1': (file_name, open(file_path, 'rb'), 'application/java-archive')
        }

        print('Uploading file to: ', data)
        response = requests.post(url, headers=headers, data=data, auth=HTTPBasicAuth(self._username, self._password),
                                 files=files)
        return response

    '''
            for now this we will create on our own, but ideally this should be created with the help of pom.xml and setup.py.
            jar - pipeline_name-1.0.jar
            wheel - pipeline_name-1.0-py3-none-any.whl
    '''

    def download_file(self, file_name: str, project_id: str, release_version: str, pipeline_name: str):
        # https: // execution.dev.cloud.prophecy.io / artifactory / repository / raw_host / ultra / demo / gamma / livy_scala.jar
        url = f'{self._base_url}/repository/{self._repository}/prophecy/artifact/{self._create_file_path(project_id, release_version, pipeline_name)}/{file_name}'
        print('Downloading file from: ', url)
        response = requests.get(url, stream=True)

        try:
            # Create a temporary directory
            tmp_dir = tempfile.mkdtemp()
            file_path = os.path.join(tmp_dir, file_name)
            with open(file_path, 'wb') as f:
                for chunk in response.iter_content(1024):
                    f.write(chunk)
            print(f"File downloaded at: {file_path}")
            return file_path
        except Exception as e:
            print("Unable to download file. HTTP response code: ", e)

            raise ArtifactDownloadFailedException(e)

    def _create_file_path(self, project_id: str, release_version: str, pipeline_name: str):
        # todo cleanup.
        return f'project_{project_id}/release_version_{release_version}/pipelines_{pipeline_name}'.replace('\\W', '_')
